from ipaddress import ip_network, ip_address
from collections import OrderedDict
from itertools import repeat
from functools import lru_cache
from collections import defaultdict
from copy import copy

import numpy as np
import pandas as pd

from suzieq.exceptions import EmptyDataframeError, PathLoopError
from .engineobj import SqPandasEngine
from suzieq.utils import expand_nxos_ifname, MAX_MTU

# TODO: What timestamp to use (arpND, mac, interface, route..)


class PathObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'path'

    def _init_dfs(self, ns, source, dest):
        """Initialize the dataframes used in this path hunt"""

        self.source = source
        self.dest = dest
        self._underlay_dfs = {}  # lpm entries for each vtep IP

        namespace = [ns]
        try:
            self._if_df = self._get_table_sqobj('interfaces') \
                .get(namespace=namespace, state='up',
                     addnl_fields=['macaddr']) \
                .explode('ipAddressList') \
                .fillna({'ipAddressList': ''}) \
                .explode('ip6AddressList') \
                .fillna({'ip6AddressList': ''}) \
                .reset_index(drop=True)
            if self._if_df.empty:
                raise EmptyDataframeError
        except (EmptyDataframeError, KeyError):
            raise EmptyDataframeError(
                f"No interface information found for {namespace}")

        # Need this in determining L2 peer
        mlag_df = self._get_table_sqobj('mlag').get(namespace=namespace)
        mlag_peers = defaultdict(str)
        mlag_peerlink = defaultdict(str)
        if not mlag_df.empty:
            peerlist = [x.tolist()
                        for x in mlag_df.groupby(by=['systemId'])['hostname']
                        .unique().tolist()]
            for peers in peerlist:
                if len(peers) > 1:
                    mlag_peers[peers[0]] = peers[1]
                    mlag_peers[peers[1]] = peers[0]
            for row in mlag_df.itertuples():
                mlag_peerlink[row.hostname] = expand_nxos_ifname(row.peerLink)
        self._mlag_peers = mlag_peers
        self._mlag_peerlink = mlag_peerlink

        try:
            # access-internal is an internal Junos route we want to
            # ignore
            self._rdf = self._get_table_sqobj('routes') \
                .lpm(namespace=namespace, address=dest) \
                .query('protocol != "access-internal"') \
                .reset_index(drop=True)
            if self._rdf.empty:
                raise EmptyDataframeError
        except (KeyError, EmptyDataframeError):
            raise EmptyDataframeError("No Routes information found for {}".
                                      format(dest))

        try:
            self._rpf_df = self._get_table_sqobj('routes') \
                .lpm(namespace=namespace, address=source) \
                .query('protocol != "access-internal"') \
                .reset_index(drop=True)
            if self._rpf_df.empty:
                raise EmptyDataframeError
        except (KeyError, EmptyDataframeError):
            raise EmptyDataframeError("No Routes information found for {}".
                                      format(source))

        # We ignore the lack of ARPND for now
        self._arpnd_df = self._get_table_sqobj('arpnd') \
            .get(namespace=namespace)
        if self._arpnd_df.empty:
            raise EmptyDataframeError(
                f"No ARPND information found for {dest}")

        # Enhance the ARPND table with the VRF field
        self._arpnd_df = self._arpnd_df.merge(
            self._if_df[['namespace', 'hostname', 'ifname', 'master']],
            left_on=['namespace', 'hostname', 'oif'],
            right_on=['namespace', 'hostname', 'ifname'], how='left') \
            .drop(columns=['ifname']) \
            .rename(columns={'master': 'vrf'}) \
            .replace({'vrf': {'': 'default'}}) \
            .query('state != "failed"') \
            .reset_index(drop=True)

        self._macsobj = self._get_table_sqobj('macs')

        if ':' in source:
            self._src_df = self._if_df[self._if_df.ip6AddressList.astype(str)
                                       .str.contains(source + "/")]
        else:
            self._src_df = self._if_df[self._if_df.ipAddressList.astype(str)
                                       .str.contains(source + "/")]

        if self._src_df.empty:
            # TODO: No host with this src addr. Is addr a local ARP entry?
            self._src_df = self._find_fhr_df(None, source)

        if self._src_df.empty:
            raise AttributeError(f"Invalid src {source}")

        if self._src_df.hostname.nunique() == 1 and len(self._src_df) > 1:
            # Multiple interfaces with the same IP address. Possible case
            # of Unnumbered interfaces. See if there's a loopback in there
            if 'loopback' in self._src_df.type.unique().tolist():
                self._src_df = self._src_df.query('type == "loopback"')

        if ':' in dest:
            self._dest_df = self._if_df[self._if_df.ip6AddressList.astype(str)
                                        .str.contains(dest + "/")]
        else:
            self._dest_df = self._if_df[self._if_df.ipAddressList.astype(str)
                                        .str.contains(dest + "/")]

        srcnet = self._src_df.ipAddressList.tolist()[0]
        if ip_address(dest) in ip_network(srcnet, strict=False):
            self.is_l2 = True
        else:
            self.is_l2 = False

        if self._dest_df.empty:
            # TODO: No host with this dest addr. Is addr a local ARP entry?
            self._dest_df = self._find_fhr_df(None, dest)

        if self._dest_df.empty:
            raise AttributeError(f"Invalid dest {dest}")

        if self._dest_df.hostname.nunique() == 1 and len(self._dest_df) > 1:
            # Multiple interfaces with the same IP address. Possible case
            # of Unnumbered interfaces. See if there's a loopback in there
            if 'loopback' in self._dest_df.type.unique().tolist():
                self._dest_df = self._dest_df.query('type == "loopback"')

        self.dest_device = self._dest_df["hostname"].unique()
        self.src_device = self._src_df["hostname"].unique()

        # Start with the source host and find its route to the destination
        if self._rdf.query(f"hostname.isin({self.src_device.tolist()})").empty:
            raise EmptyDataframeError(f"No routes found for {self.src_device}")

    def _get_vrf(self, hostname: str, ifname: str, addr: str) -> str:
        """Determine the VRF given either the ifname or ipaddr"""
        vrf = ''
        if addr and ':' in addr:
            ipvers = 6
        else:
            ipvers = 4
        if ifname:
            iifdf = self._if_df[(self._if_df["hostname"] == hostname) &
                                (self._if_df["ifname"] == ifname)]
        else:
            # TODO: Add support for IPv6 here
            addr = addr + '/'
            if ipvers == 6:
                iifdf = self._if_df.query(
                    f'hostname=="{hostname}" and '
                    f'ip6AddressList.str.startswith("{addr}")')
            else:
                iifdf = self._if_df.query(
                    f'hostname=="{hostname}" and '
                    f'ipAddressList.str.startswith("{addr}")')
        if not iifdf.empty and iifdf.iloc[0].master == "bridge":
            # OK, find the SVI associated with this interface
            if addr and ipvers == 6:
                iifdf = self._if_df.query(
                    f'hostname=="{hostname}" and '
                    f'ip6AddressList.str.startswith("{addr}")')
            elif addr and ipvers == 4:
                iifdf = self._if_df.query(
                    f'hostname=="{hostname}" and '
                    f'ipAddressList.str.startswith("{addr}")')
            else:
                # No address, but a bridge interface as the master
                if iifdf.iloc[0]['vlan']:
                    # Check if there's an SVI, assuming format is vlan*
                    # TODO: Handle trunk
                    vlan = iifdf.iloc[0]['vlan']
                    vdf = self._if_df.query(f'hostname=="{hostname}" and '
                                            f'(ifname=="vlan{vlan}" or '
                                            f'ifname=="Vlan{vlan}")')
                    if not vdf.empty:
                        vrf = vdf.iloc[0].master.strip()

        if not iifdf.empty and not vrf:
            vrf = iifdf.iloc[0].master.strip()
            if vrf == "bridge":
                vrf = ''

        return vrf

    def _find_fhr_df(self, device: str, ip: str) -> pd.DataFrame:
        """Find Firstt Hop Router's iface DF for a given IP and device.
        The logic in finding the next hop router is:
          find the arp table entry corresponding to the IP provided;
          if this result is empty or the MAC addr is not unique:
             return null;
          extract unique MAC and search MAC table for this MAC;
          if the MAC table search is empty:
             return null;
          concat the dataframe using every row of the MAC DF to find if_df
          Return the concat if_df as FHR
        """
        fhr_df = pd.DataFrame()
        if not ip or self._arpnd_df.empty:
            return fhr_df

        rslt_df = self._arpnd_df.query(
            f'ipAddress=="{ip}" and not remote')

        if rslt_df.empty:
            return fhr_df

        # If we have more than one MAC addr, its hard to know which one
        uniq_mac = rslt_df['macaddr'].unique().tolist()
        if len(uniq_mac) != 1:
            return fhr_df

        macdf = self._macsobj.get(namespace=[rslt_df.iloc[0].namespace],
                                  macaddr=[uniq_mac[0]], localOnly=True)
        if not macdf.empty:
            ign_ifs = ["bridge", "Vxlan1"]
            if device:
                macdf = macdf.query(
                    f'vlan != 0 and not oif.isin({ign_ifs}) and '
                    f' hostname != "{self._mlag_peers[device]}"')
            else:
                macdf = macdf.query(f'vlan != 0 and not oif.isin({ign_ifs})')

        for row in rslt_df.itertuples():
            mac_entry = macdf.query(f'hostname == "{row.hostname}"')
            if not mac_entry.empty:
                # for row in macdf.iterrows():
                idf = self._if_df.query(
                    f'hostname=="{row.hostname}" and '
                    f'ifname=="{mac_entry.oif.iloc[0]}"').copy()

                # We need to replace the VLAN in the if_df with what
                # is obtained from the MAC because of trunk ports.
                if idf.empty:
                    continue
                idf.at[idf.index, 'vlan'] = mac_entry.vlan.iloc[0]
                if (idf.master == 'bridge').all():
                    idf.at[idf.index, 'ipAddressList'] = f'{ip}/32'
                    # Assuming the VRF is identical across multiple entries
                    idf.at[idf.index, 'master'] = rslt_df.iloc[0].vrf
            else:
                idf = self._if_df.query(f'hostname=="{row.hostname}" and '
                                        f'ifname=="{row.oif}" and '
                                        f'type=="vlan"').copy()
                if idf.empty:
                    continue
            if fhr_df.empty:
                fhr_df = idf
            else:
                fhr_df = pd.concat([fhr_df, idf])
        return fhr_df

    def _get_if_vlan(self, device: str, ifname: str) -> int:
        oif_df = self._if_df[(self._if_df["hostname"] == device) &
                             (self._if_df["ifname"] == ifname)]

        if oif_df.empty:
            return []

        return oif_df.iloc[0]["vlan"]

    def _get_l2_nexthop(self, device: str, vrf: str, dest: str,
                        macaddr: str, protocol: str) -> list:
        """Get the bridged/tunnel nexthops
        We're passing protocol because we need to keep the return
        match the other get nexthop function returns. We don't really
        use protocol or determine it.
        """

        if self._arpnd_df.empty:
            return []

        rslt = self._arpnd_df[(self._arpnd_df['hostname'] == device) &
                              (self._arpnd_df['ipAddress'] == dest)]
        # the end of knowledge
        if rslt.empty:
            # Check if we have an EVPN entry as a route (symmetric routing)
            rslt = self._rdf.query(
                f'hostname == "{device}" and vrf == "{vrf}"')
            if not rslt.empty:
                # Check that we have a host route at this point
                ipvers = 6 if ':' in dest else 4
                if ((ipvers == 4 and rslt.iloc[0].prefix == f'{dest}/32') or
                        (ipvers == 6 and rslt.iloc[0].prefix == f'{dest}/128')):
                    overlay = rslt.iloc[0].nexthopIps[0]
                    if not overlay:
                        raise AttributeError('missing overlay')
                    return self._get_underlay_nexthop(device, [overlay],
                                                      ['default'], True)
            return []

        if rslt.empty:
            return []

        if not macaddr:
            macaddr = rslt.iloc[0]["macaddr"]
        oif = rslt.iloc[0]["oif"]
        # This is to handle the VRR interface on Cumulus Linux machines
        if oif.endswith("-v0"):
            oif = oif.split("-v0")[0]

        vlan = self._get_if_vlan(device, oif)

        if macaddr:
            mac_df = self._macsobj.get(namespace=[rslt.iloc[0]['namespace']],
                                       hostname=[device], macaddr=[macaddr],
                                       vlan=[str(vlan)])

            if mac_df.empty:
                # On servers there's no bridge and thus no MAC table entry
                return [('', rslt.iloc[0].oif, False, protocol == "l2",
                         protocol, np.nan)]

            overlay = mac_df.iloc[0].remoteVtepIp or False
            if not overlay:
                return ([('', mac_df.iloc[0].oif, overlay, protocol == "l2",
                          protocol,
                          mac_df.iloc[0].timestamp)])
            else:
                # We assume the default VRF as the underlay. Can this change?
                return self._get_underlay_nexthop(device, [overlay],
                                                  ['default'], True)
        return []

    def _get_underlay_nexthop(self, hostname: str, vtep_list: list,
                              vrf_list: list, is_overlay: bool) -> pd.DataFrame:
        """Return the underlay nexthop given the Vtep and VRF"""

        # WARNING: This function is incomplete right now
        result = []

        if len(set(vrf_list)) != 1:
            # VTEP underlay can only be in a single VRF
            return result

        vrf = vrf_list[0].split(':')[-1]
        for vtep in vtep_list:
            if vtep not in self._underlay_dfs:
                if not vtep:
                    raise AttributeError(
                        f'false vtep {vtep_list}: vrf {vrf_list}')
                self._underlay_dfs[vtep] = self._get_table_sqobj('routes') \
                    .lpm(namespace=[self.namespace],
                         address=[vtep], vrf=[vrf])
            vtep_df = self._underlay_dfs[vtep]
            rslt = vtep_df.query(
                f'hostname == "{hostname}" and vrf == "{vrf}"')
            if not rslt.empty:
                if is_overlay:
                    intres = zip(rslt.nexthopIps.iloc[0].tolist(),
                                 rslt.oifs.iloc[0].tolist(),
                                 repeat(vtep), repeat(is_overlay),
                                 repeat(rslt.protocol.iloc[0]),
                                 repeat(rslt.timestamp.iloc[0])
                                 )
                elif rslt.protocol.iloc[0] == 'direct':
                    intres = zip([vtep], [rslt.oifs.iloc[0][0]], [False], [False],
                                 [rslt.protocol.iloc[0]], [rslt.timestamp.iloc[0]])

                else:
                    intres = zip(rslt.nexthopIps.iloc[0].tolist(),
                                 rslt.oifs.iloc[0].tolist(),
                                 repeat(False), repeat(is_overlay),
                                 repeat(rslt.protocol.iloc[0]),
                                 repeat(rslt.timestamp.iloc[0])
                                 )

                result.extend(list(intres))

        return result

    def _get_nexthops(self, device: str, vrf: str, dest: str, is_l2: bool,
                      vtep: str, macaddr: str) -> list:
        """Get nexthops (oif + IP + overlay) or just oif for given host/vrf.

        The overlay is a bit indicating we're getting into overlay or not.
        """

        if is_l2:
            if vtep:
                return self._get_underlay_nexthop(device, [vtep], [vrf], True)
            else:
                return self._get_l2_nexthop(device, vrf, dest, macaddr, 'l2')

        rslt = self._rdf.query(f'hostname == "{device}" and vrf == "{vrf}"')
        # The following condition is checking that we have a pure L3 nexthop or
        # the start of an underlay route. if its a pure L3 route, the nexthopIp
        # is not empty OR the protocol is not hmm--NXOS' host mobility
        # protocol--which also puts a final subnet route with EVPN also as an L3
        # route.
        if not rslt.empty and (len(rslt.nexthopIps.iloc[0]) != 0 and
                               rslt.nexthopIps.iloc[0][0] != '') and (
                                   rslt.protocol.iloc[0] != 'hmm'):
            # Handle overlay routes given how NXOS programs its FIB
            if rslt.oifs.explode().str.startswith('_nexthopVrf:').any():
                # We need to find the true NHIP
                if rslt.oifs.iloc[0].tolist() == [False]:
                    raise AttributeError('mising vtep')
                return self._get_underlay_nexthop(
                    device,
                    rslt.nexthopIps.iloc[0].tolist(),
                    rslt.oifs.iloc[0].tolist(), True)
            elif not rslt.empty and (len(rslt.nexthopIps.iloc[0]) != 0 and
                                     rslt.nexthopIps.iloc[0][0] != '') and (
                                         not rslt.oifs.iloc[0].tolist()):
                # NXOS and recursive route handling
                if rslt.ofs.iloc[0].tolist() == [False]:
                    raise AttributeError('mising vtep recursive route')
                return self._get_underlay_nexthop(
                    device, rslt.nexthopIps.iloc[0].tolist(), [vrf], False)
            return zip(rslt.nexthopIps.iloc[0].tolist(),
                       rslt.oifs.iloc[0].tolist(),
                       repeat(False), repeat(False),
                       repeat(rslt.protocol.iloc[0]),
                       repeat(rslt.timestamp.iloc[0]))

        # We've either reached the end of routing or the end of knowledge
        # Look for L2 nexthop
        if rslt.empty:
            protocol = ''
        else:
            protocol = rslt.protocol.iloc[0]
        return self._get_l2_nexthop(device, vrf, dest, None, protocol)

    @ lru_cache(maxsize=256)
    def _get_nh_with_peer(self, device: str, vrf: str, dest: str, is_l2: bool,
                          vtep_ip: str, macaddr: str) -> list:
        """Get the nexthops & peer node for each nexthop for a given device/vrf
        This uses the cached route lpm DF to get the nexthops. It
        also handles vlan subinterfaces, MLAG and plain bonds to get the
        valid nexthop peers.

        :param device: devicename to query in lpm DF for nexthop
        :type device: str

        :param vrf: VRF to qualify lpm DF for nexthop
        :type vrf: str

        :param dest: Destination IP we're searching forever, needed for arp/nd
        :type dest: str

        :param is_l2: If this is an L2 lookup
        :type is_l2: bool

        :param vtep_ip: VTEP IP address, can be empty string if not in underlay
        :type vtep_ip: str

        :param macaddr: Look for this MAC addr, not dest, as this is L2hop
        :type macaddr: str

        :rtype: list
        :return:
        list of tuples where each tuple is (oif, peerdevice, peerif, overlay)
        """

        nexthops = []

        # TODO: Can we have ECMP nexthops, one with NH IP and one without?
        nexthop_list = self._get_nexthops(
            device, vrf, dest, is_l2, vtep_ip, macaddr)
        new_nexthop_list = []
        # Convert each OIF into its actual physical list
        is_l2 = False
        for nhip, iface, overlay, is_l2, protocol, timestamp in nexthop_list:
            if macaddr and is_l2 and not overlay:
                new_nexthop_list.append((nhip, iface, overlay, is_l2,
                                         protocol, timestamp))
                continue

            if (not nhip or nhip == 'None') and iface:
                addr = dest
            else:
                addr = nhip
            if iface.endswith('-v0'):
                # Replace Cumulus' VRR entry with actual SVI
                iface = iface.split('-v0')[0]
            # This first pass is to handle Cumulus symmetric EVPN routes
            arpdf = self._arpnd_df.query(f'hostname=="{device}" and '
                                         f'ipAddress=="{addr}" and '
                                         f'oif=="{iface}"')
            if not arpdf.empty and arpdf.remote.all():
                macdf = self._macsobj.get(namespace=[self.namespace],
                                          hostname=[device],
                                          macaddr=[arpdf.iloc[0].macaddr])
                if not macdf.empty and macdf.remoteVtepIp.all():
                    overlay = macdf.iloc[0].remoteVtepIp

                # you don't always have an overlay if the network is transitioning
                if overlay:
                    underlay_nh = self._get_underlay_nexthop(device, [overlay],
                                                             ['default'], True)
                    new_nexthop_list.extend(underlay_nh)
            else:
                new_nexthop_list.append((nhip, iface, overlay, is_l2,
                                         protocol, timestamp))

        on_src_node = device in self.src_device
        for (nhip, iface, overlay, is_l2, protocol,
             timestamp) in new_nexthop_list:
            df = pd.DataFrame()
            errormsg = ''
            if is_l2 and macaddr and not overlay:
                if (not nhip or nhip == 'None') and iface:
                    addr = dest + '/'
                else:
                    addr = nhip + '/'
                df = self._if_df.query(
                    f'macaddr=="{macaddr}" '
                    f'and type != "bond_slave" '
                    f'and state != "down" '
                    f'and ipAddressList.str.startswith("{addr}")')
                if df.empty:
                    df = self._if_df.query(
                        f'ipAddressList.str.startswith("{addr}") and '
                        f'type !="bond_slave"')
                    if df.empty:
                        continue
            else:
                if not nhip:
                    nhip = dest
                arpdf = self._arpnd_df.query(f'hostname=="{device}" and '
                                             f'ipAddress=="{nhip}" and '
                                             f'oif=="{iface}"')
                if not arpdf.empty:
                    addr = nhip + '/'
                    df = self._if_df.query(
                        f'macaddr=="{arpdf.iloc[0].macaddr}" '
                        f'and type != "bond_slave" '
                        f'and state != "down" '
                        f'and ipAddressList.str.startswith("{addr}")')
                    if df.empty:
                        # In case of L2 interfaces as the nexthop, there'll be
                        # no IP address on the interface with matching NHIP.
                        df = self._if_df.query(
                            f'macaddr=="{arpdf.iloc[0].macaddr}" '
                            f'and type != "bond_slave"')
                elif protocol == 'direct':
                    continue
                nhip_df = self._if_df.query(
                    f'ipAddressList.str.startswith("{nhip}/") and '
                    f'type != "bond_slave" and hostname != "{device}"')
                if df.empty and not nhip_df.empty:
                    df = nhip_df
                elif on_src_node and not df.empty and not nhip_df.empty:
                    if ((df.hostname.unique().tolist() !=
                         nhip_df.hostname.unique().tolist()) and
                        (df.macaddr.unique().tolist() !=
                         nhip_df.macaddr.unique().tolist())):
                        errormsg = 'Possible anycast IP without anycast MAC'
                        is_l2 = True

                if df.empty:
                    continue

            # In case of centralized EVPN, its possible to find the NHIP on
            # unconnected devices if this is still the source device i.e.
            # server. If so, find FHR to get the real connected dev
            if device in self.src_device:
                check_fhr = self.source
            elif is_l2 and (df.iloc[0].hostname in self.dest_device):
                check_fhr = self.dest
            else:
                check_fhr = None

            if check_fhr:
                fhr_df = self._find_fhr_df(device, check_fhr)
                if not fhr_df.empty:
                    fhr_hosts = set(fhr_df['hostname'].tolist())

                    if check_fhr != self.dest and device not in fhr_hosts:
                        # Avoid looping everytime we hit the dest device
                        # We only need to do it once
                        df = df.query(f'hostname.isin({list(fhr_hosts)})')
                        if df.empty:
                            # The L3 nexthop is not the next L2 hop
                            df = fhr_df
                            is_l2 = True
                            # But what we're going to be forwarding on when
                            # we hit the next L2 hop is the MAC addr from the
                            # ARP entry, and so set the macaddr to that for now
                            if not arpdf.empty:
                                macaddr = arpdf.macaddr.unique().tolist()[0]
                        else:
                            # Fix the incoming interface
                            df = df.merge(fhr_df[['namespace', 'hostname',
                                                  'ifname']],
                                          on=['namespace', 'hostname'],
                                          how='outer') \
                                .fillna(method='ffill') \
                                .drop(columns=['ifname_x']) \
                                .rename(columns={'ifname_y': 'ifname'})

            # In case of some NOS such as NXOS with OSPF unnumbered, multiple
            # interfaces from the same device have the same IP and MAC. This
            # needs to be filtered to the precise interface. We do this by
            # matching the IP/MAC of this device's OIF with the ARP/ND table
            # on the nexthop device
            if (df.hostname.nunique() == 1) and (df.ifname.nunique() > 1):
                oif_df = self._if_df.query(f'hostname=="{device}" and '
                                           f' ifname=="{iface}"')
                if not oif_df.empty and oif_df.ipAddressList.iloc[0]:
                    revip = oif_df.ipAddressList.iloc[0].split('/')[0]
                    revvrf = "default" if overlay else vrf
                    revarp_df = self._arpnd_df.query(
                        f'hostname=="{df.hostname.unique()[0]}" and '
                        f'ipAddress=="{revip}" and vrf=="{revvrf}" and '
                        'state!="failed"')
                    if not revarp_df.empty:
                        df = df.query(f'ifname == "{revarp_df.oif.iloc[0]}"')
            df.apply(lambda x, nexthops:
                     nexthops.append((iface, x['hostname'],
                                      x['ifname'],  overlay,
                                      is_l2, nhip,
                                      macaddr or x.macaddr, protocol, errormsg,
                                      timestamp))
                     if (x['namespace'] in self.namespace) else None,
                     args=(nexthops,), axis=1)

        if not nexthops and is_l2:
            return [(None, None, None, False, is_l2, None, None, "l2", "",
                     np.nan)]

        return sorted(nexthops, key=lambda x: x[1])

    def get(self, **kwargs) -> pd.DataFrame:
        """return a pandas dataframe with the paths between src and dest
        :param kwargs:
        :return:
        :rtype: pd.DataFrame
        """

        if not self.ctxt.engine:
            raise AttributeError(
                "Specify an analysis engine using set engine " "command"
            )

        namespaces = kwargs.get("namespace", self.ctxt.namespace)
        if not namespaces:
            raise AttributeError("Must specify namespace to run the trace in")

        self.namespace = namespaces[0]
        src = kwargs.get("source", None)
        dest = kwargs.get("dest", None)
        dvrf = kwargs.get("vrf", "")

        if not src or not dest:
            raise AttributeError("Must specify trace source and dest")

        srcvers = ip_network(src, strict=False)._version
        dstvers = ip_network(dest, strict=False)._version
        if srcvers != dstvers:
            raise AttributeError(
                "Source and Dest MUST belong to same address familt")
        # All exceptions in the initial data gathering will happen in this init
        # After this, at least we know we have the data to work on
        self._init_dfs(self.namespace, src, dest)

        devices_iifs = OrderedDict()
        src_mtu = None
        for i in range(len(self._src_df)):
            item = self._src_df.iloc[i]
            devices_iifs[f'{item.hostname}/'] = {
                "iif": item["ifname"],
                "mtu": item["mtu"],
                "outMtu": item["mtu"],
                "overlay": '',
                "protocol": '',
                "error": [],
                "lookup": dest,
                "macaddr": None,
                "vrf": item['master'],
                'mtuMatch': True,  # Its the first node
                "is_l2": self.is_l2,
                "nhip": '',
                "overlay_nhip": '',
                "oif": item["ifname"],
                "timestamp": item["timestamp"],
                "l3_visited_devices": set(),
                "l2_visited_devices": set()
            }
            if src_mtu is None or (item.get('mtu', 0) < src_mtu):
                src_mtu = item.get('mtu', 0)
        if not dvrf:
            dvrf = item['master']
        if not dvrf:
            dvrf = "default"

        dest_device_iifs = OrderedDict()
        for i in range(len(self._dest_df)):
            item = self._dest_df.iloc[i]
            error = []
            dest_device_iifs[f'{item.hostname}/'] = {
                "iif": '',
                "vrf": item["master"] or "default",
                "mtu": item["mtu"],
                "outMtu": item["mtu"],
                "macaddr": None,
                "error": error,
                "overlay": '',
                "is_l2": False,
                "overlay_nhip": '',
                "oif": item["ifname"],
                "protocol": '',
                "lookup": '',
                "timestamp": item["timestamp"],
            }

        final_paths = []
        paths = []
        on_src_node = True

        # The logic is to loop through the nexthops till you reach the dest
        # device The topmost while is this looping. The next loop within handles
        # one nexthop at a time.The paths are constructed as a list of lists,
        # where each element of the outermost loop is one complete path and
        # each inner list represents one hop in that path. Each hop is the
        # list of devicename and incoming interface. loops are detected by
        # ensuring that no device is visited twice in the same VRF. The VRF
        # qualification is required to ensure packets coming back from a
        # firewall or load balancer are not tagged as duplicates.
        while devices_iifs:
            nextdevices_iifs = OrderedDict()
            newpaths = []
            revdf_check = True

            for devkey in devices_iifs:
                device = devkey.split('/')[0]
                iif = devices_iifs[devkey]["iif"]
                devvrf = devices_iifs[devkey]["vrf"]
                ioverlay = devices_iifs[devkey]["overlay"]
                macaddr = devices_iifs[devkey]['macaddr']
                l3_visited_devices = devices_iifs[devkey]['l3_visited_devices']
                l2_visited_devices = devices_iifs[devkey]['l2_visited_devices']
                is_l2 = devices_iifs[devkey]['is_l2']

                # We've reached the destination, so stop this loop
                destdevkey = f'{device}/'
                if destdevkey in dest_device_iifs:
                    if revdf_check:
                        vrfchk = dest_device_iifs[destdevkey]["vrf"]
                        rev_df = self._rpf_df.query(
                            f'hostname == "{device}" and vrf == "{vrfchk}"')
                        if rev_df.empty:
                            dest_device_iifs[destdevkey]['error'] \
                                .append('no reverse path')
                        revdf_check = False
                    pdev1 = devkey.split('/')[1]
                    for x in paths:
                        pdev2 = list(x[-1].keys())[0].split('/')[0]
                        if (pdev1 != pdev2) and not on_src_node:
                            continue
                        copy_dest = copy(dest_device_iifs[destdevkey])
                        if not on_src_node:
                            copy_dest['oif'] = devices_iifs[devkey]['oif']
                        copy_dest['iif'] = iif
                        copy_dest['mtu'] = devices_iifs[devkey]['mtu']
                        dst_mtu = copy_dest.get('outMtu', 0)
                        if dst_mtu > MAX_MTU:
                            dst_mtu = copy_dest.get('mtu', 0)
                        if dst_mtu != src_mtu:
                            if 'Dst MTU != Src MTU' not in copy_dest['error']:
                                copy_dest['error'].append('Dst MTU != Src MTU')
                        # This is weird because we have no room to store the
                        # prev hop's outgoing IIF MTU on the last hop
                        copy_dest['outMtu'] = \
                            f'{devices_iifs[devkey]["outMtu"]}/' \
                            f'{dest_device_iifs[destdevkey]["outMtu"]}'
                        copy_dest['is_l2'] = is_l2
                        if not is_l2:
                            copy_dest['nhip'] = dest
                        copy_dest['mtuMatch'] = devices_iifs[devkey]['mtuMatch']
                        z = x + [OrderedDict(
                            {destdevkey: copy_dest})]
                        if z not in final_paths:
                            final_paths.append(z)

                    if not paths:
                        copy_dest = copy(dest_device_iifs[destdevkey])
                        # This is because the src and dst nodes are the same
                        # So ignore the OIF coming out of devices_iifs
                        copy_dest['iif'] = iif
                        copy_dest['mtu'] = devices_iifs[devkey]['mtu']
                        dst_mtu = copy_dest.get('outMtu', 0)
                        if dst_mtu > MAX_MTU:
                            dst_mtu = copy_dest.get('mtu', 0)
                        if dst_mtu != src_mtu:
                            if 'Dst MTU != Src MTU' not in copy_dest['error']:
                                copy_dest['error'].append('Dst MTU != Src MTU')
                        # This is weird because we have no room to store the
                        # prev hop's outgoing IIF MTU on the last hop
                        copy_dest['outMtu'] = \
                            f'{devices_iifs[devkey]["outMtu"]}/' \
                            f'{dest_device_iifs[destdevkey]["outMtu"]}'
                        copy_dest['is_l2'] = is_l2
                        if not is_l2:
                            copy_dest['nhip'] = dest
                        copy_dest['mtuMatch'] = devices_iifs[devkey]['mtuMatch']

                        paths = [[OrderedDict({destdevkey: copy_dest})]]
                    continue

                newdevices_iifs = {}  # NHs from this NH to add to the next round
                end_overlay = True
                if is_l2 or ioverlay:
                    if ioverlay:
                        ndst = ioverlay
                    else:
                        ndst = devices_iifs[devkey].get('nhip', None)
                    # Check if this is the end of the L2 path or overlay
                    nhdf = self._if_df.query(f'hostname=="{device}"')
                    if not nhdf.empty:
                        if srcvers == 4:
                            nhdf = nhdf.query(
                                f'ipAddressList.str.contains("{ndst}")')
                        else:
                            nhdf = nhdf.query(
                                f'ip6AddressList.str.contains("{ndst}")')
                    if not nhdf.empty:
                        ifmac = nhdf.macaddr.unique().tolist()
                        if (not macaddr) or (macaddr in ifmac):
                            ndst = dest
                            is_l2 = self.is_l2
                            ioverlay = ''
                            l2_visited_devices = set()
                    elif ioverlay:
                        end_overlay = False
                else:
                    ndst = dest
                if not ndst:
                    ndst = dest

                if not end_overlay:
                    ivrf = 'default'
                elif end_overlay and devices_iifs[devkey]['overlay_nhip']:
                    ivrf = self._get_vrf(device, '',
                                         devices_iifs[devkey]['overlay_nhip'])
                    if not ivrf:
                        ivrf = devvrf
                elif (not devices_iifs[devkey]['nhip'] or
                      (devices_iifs[devkey]['nhip'] == "169.254.0.1")):
                    ivrf = self._get_vrf(device, iif, '')
                else:
                    ivrf = self._get_vrf(device, '',
                                         devices_iifs[devkey]['nhip'])
                    if not ivrf:
                        ivrf = self._get_vrf(device, iif, '')

                if not ivrf:
                    ivrf = dvrf

                skey = device + ivrf
                if is_l2:
                    if skey in l2_visited_devices:
                        # This is a loop
                        if ioverlay:
                            devices_iifs[devkey]['error'] \
                                .append("Loop in underlay")
                        else:
                            devices_iifs[devkey]['error'] \
                                .append("L2 Loop detected")
                        for x in paths:
                            z = x + [OrderedDict({devkey:
                                                  devices_iifs[devkey]})]
                            if z not in final_paths:
                                final_paths.append(z)
                        continue
                    else:
                        l2_visited_devices.add(skey)
                else:
                    if skey in l3_visited_devices:
                        devices_iifs[devkey]['error'].append("L3 loop")
                        for x in paths:
                            z = x + [OrderedDict({devkey:
                                                  devices_iifs[devkey]})]
                            if z not in final_paths:
                                final_paths.append(z)
                        continue
                    else:
                        l3_visited_devices.add(skey)

                devices_iifs[devkey]['vrf'] = ivrf
                rt_ts = None
                if not (is_l2 or ioverlay):
                    rslt = self._rdf.query('hostname == "{}" and vrf == "{}"'
                                           .format(device, ivrf))
                    if not rslt.empty:
                        devices_iifs[devkey]['timestamp'] = rslt.timestamp.iloc[0]
                        devices_iifs[devkey]['protocol'] = rslt.protocol.iloc[0]
                        devices_iifs[devkey]['lookup'] = rslt.prefix.iloc[0]

                        rev_df = self._rpf_df.query(
                            f'hostname == "{device}" and vrf == "{ivrf}"')
                        if rev_df.empty and not on_src_node:
                            devices_iifs[devkey]['error'] \
                                .append('no reverse path')
                    else:
                        devices_iifs[devkey]['lookup'] = ''

                elif macaddr:
                    devices_iifs[devkey]['lookup'] = ''
                else:
                    devices_iifs[devkey]['lookup'] = ndst

                for i, nexthop in enumerate(self._get_nh_with_peer(
                        device, ivrf, ndst, is_l2, ioverlay, macaddr)):
                    error = []
                    (iface, peer_device, peer_if, overlay, is_l2,
                     nhip, macaddr, protocol, errmsg, timestamp) = nexthop
                    if not devices_iifs[devkey].get('protocol', ''):
                        devices_iifs[devkey]['protocol'] = protocol
                    if not rt_ts:
                        devices_iifs[devkey]['timestamp'] = timestamp
                    if errmsg and errmsg not in devices_iifs[devkey]['error']:
                        devices_iifs[devkey]['error'].append(errmsg)

                    if iface is not None:
                        if iface.startswith('vPC Peer'):
                            iface = self._mlag_peerlink[device]
                        elif iface.startswith('sup-eth1'):
                            iface = 'loopback0'
                        if peer_if.startswith('vPC Peer'):
                            peer_if = self._mlag_peerlink[peer_device]
                        elif peer_if.startswith('sup-eth1'):
                            peer_if = 'loopback0'
                        in_mtu = self._if_df[
                            (self._if_df["hostname"] == peer_device) &
                            (self._if_df["ifname"] == peer_if)
                        ].iloc[-1].mtu
                        out_mtu = self._if_df[
                            (self._if_df["hostname"] == device) &
                            (self._if_df["ifname"] == iface)
                        ].iloc[-1].mtu
                        if on_src_node and src_mtu > MAX_MTU:
                            src_mtu = out_mtu
                        mtu_match = in_mtu == out_mtu
                        if (in_mtu < src_mtu) or (out_mtu < src_mtu):
                            error.append('Hop MTU < Src Mtu')
                        if not end_overlay:
                            overlay = ioverlay
                            vrf = devvrf
                        else:
                            vrf = ivrf
                        if not nhip:
                            nhip = ndst
                        if overlay and not ioverlay:
                            overlay_nhip = ndst
                        elif not end_overlay:
                            overlay_nhip = devices_iifs[devkey]['overlay_nhip']
                        else:
                            overlay_nhip = ''
                        if overlay or not is_l2:
                            # We don't need to track MACaddr if its not a pure
                            # (non-overlay) L2 hop
                            macaddr = None

                        newdevices_iifs[f'{peer_device}/{device}'] = {
                            "iif": peer_if,
                            "vrf": vrf,
                            "macaddr": macaddr,
                            "overlay": overlay,
                            "mtu": in_mtu,
                            "outMtu": out_mtu,  # prev hop's outMTU, fixed ltr
                            "mtuMatch": mtu_match,
                            "is_l2": is_l2,
                            "nhip": nhip,
                            "oif": iface,
                            "error": error,
                            "overlay_nhip": overlay_nhip,
                            'l3_visited_devices': l3_visited_devices.copy(),
                            'l2_visited_devices': l2_visited_devices.copy()
                        }

                if not paths:
                    newpaths.append(
                        [OrderedDict({devkey: devices_iifs[devkey]})])

                # matching to attach the hop to the appropriate path
                pdev1 = devkey.split('/')[1]
                for x in paths:
                    xkey = list(x[-1].keys())[0]
                    pdev2 = xkey.split('/')[0]
                    if pdev1 != pdev2:
                        continue
                    z = x + [OrderedDict({devkey: devices_iifs[devkey]})]
                    if z not in newpaths:
                        newpaths.append(z)

                for x in newdevices_iifs:
                    if x not in nextdevices_iifs:
                        nextdevices_iifs[x] = newdevices_iifs[x]

            if newpaths:
                paths = newpaths

            devices_iifs = nextdevices_iifs
            on_src_node = False

        if not final_paths:
            # This occurs when a path traversal terminates due to an error such
            # as loop detected
            final_paths = paths
        return self._path_cons_result(final_paths)

    def _path_cons_result(self, paths):
        df_plist = []
        prev_hop = hop = None
        for i, path in enumerate(paths):
            if prev_hop:
                # Taking advantage of python's shallow copy, that this
                # also changes what's in df_plist
                prev_hop['oif'] = self._dest_df.iloc[0]['ifname']
                if isinstance(prev_hop['outMtu'], str) and '/' in prev_hop['outMtu']:
                    prev_hop['outMtu'] = int(prev_hop['outMtu'].split('/')[1])
                prev_hop['isL2'] = False
                prev_hop['nexthopIp'] = ''
                prev_hop['vtepLookup'] = ''
                prev_hop = None
            for j, ele in enumerate(path):
                item = list(ele)[0]
                if ele[item]['overlay']:
                    overlay = True
                else:
                    overlay = False
                lookup = ele[item].get("lookup", "")

                hop = {
                    "pathid": i + 1,
                    "hopCount": j,
                    "namespace": self.namespace,
                    "hostname": item.split('/')[0],
                    "iif": ele[item]["iif"],
                    "oif": ele[item]['oif'],
                    "vrf": ele[item]["vrf"],
                    "isL2": ele[item].get("is_l2", False),
                    "overlay": overlay,
                    "mtuMatch": ele[item].get("mtuMatch", np.nan),
                    "inMtu": ele[item].get("mtu", 0),
                    "outMtu": ele[item].get("outMtu", 0),
                    "protocol": ele[item].get('protocol', ''),
                    "ipLookup": lookup,
                    "vtepLookup": "",
                    "macLookup": "",
                    "nexthopIp": ele[item].get('nhip', ''),
                    "error": ', '.join(ele[item].get('error', [])),
                    "timestamp": ele[item].get("timestamp", np.nan)
                }
                df_plist.append(hop)
                # Update some of the edge info in the prev rowerr
                # Some path specific info cannot be populated in a hop
                # because of the algorithm. So, we save the info associated
                # with each hop of the path in the nexthop and have to do
                # this logic to populate the prev hop
                if prev_hop:
                    # Taking advantage of python's shallow copy, that this
                    # also changes what's in df_plist
                    prev_hop['nexthopIp'] = hop['nexthopIp']
                    if overlay:
                        prev_hop['vtepLookup'] = ele[item]["overlay"]
                    elif hop["isL2"]:
                        prev_hop["macLookup"] = ele[item]["macaddr"]
                    prev_hop['isL2'] = hop['isL2']
                    prev_hop['oif'] = hop['oif']
                    if isinstance(hop['outMtu'], str) and '/' in hop['outMtu']:
                        prev_hop['outMtu'] = int(hop['outMtu'].split('/')[0])
                    else:
                        prev_hop['outMtu'] = hop['outMtu']
                prev_hop = hop
        if prev_hop:
            # Taking advantage of python's shallow copy, that this
            # also changes what's in df_plist
            prev_hop['oif'] = self._dest_df.iloc[0]['ifname']
            if isinstance(prev_hop['outMtu'], str) and '/' in prev_hop['outMtu']:
                prev_hop['outMtu'] = int(prev_hop['outMtu'].split('/')[1])
            prev_hop['isL2'] = False
            prev_hop['nexthopIp'] = ''
            prev_hop['vtepLookup'] = ''
        paths_df = pd.DataFrame(df_plist)
        if not paths_df.empty and not any(paths_df.error):
            paths_df.drop(columns=['error'], inplace=True)
        return paths_df.drop_duplicates()

    def summarize(self, **kwargs):
        """return a pandas dataframe summarizing the path info between src/dest

        A copy of this function is in gui_path.py for reasons of efficiency
        :param kwargs:
        :return:
        :rtype: pd.DataFrame
        """

        path_df = self.get(**kwargs)

        if path_df.empty:
            return pd.DataFrame()

        namespace = self.namespace
        ns = {}
        ns[namespace] = {}

        perhopEcmp = path_df.query('hopCount != 0') \
                            .groupby(by=['hopCount'])['hostname']
        ns[namespace]['totalPaths'] = path_df['pathid'].max()
        ns[namespace]['perHopEcmp'] = perhopEcmp.nunique().tolist()
        ns[namespace]['maxPathLength'] = path_df.groupby(by=['pathid'])[
            'hopCount'].max().max()
        ns[namespace]['avgPathLength'] = path_df.groupby(by=['pathid'])[
            'hopCount'].max().mean()
        ns[namespace]['uniqueDevices'] = path_df['hostname'].nunique()
        ns[namespace]['mtuMismatch'] = not all(path_df['mtuMatch'])
        ns[namespace]['usesOverlay'] = any(path_df['overlay'])
        ns[namespace]['pathMtu'] = min(path_df.query('iif != "lo"')['inMtu'].min(),
                                       path_df.query('iif != "lo"')['outMtu'].min())

        summary_fields = ['totalPaths', 'perHopEcmp', 'maxPathLength',
                          'avgPathLength', 'uniqueDevices', 'pathMtu',
                          'usesOverlay', 'mtuMismatch']
        return pd.DataFrame(ns).reindex(summary_fields, axis=0) \
                               .convert_dtypes()



from typing import List
import pandas as pd
import numpy as np
from .engineobj import SqPandasEngine
from suzieq.utils import convert_macaddr_format_to_colon


class NetworkObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'network'

    def get(self, **kwargs):
        """Get the information requested"""

        view = kwargs.get('view', self.iobj.view)
        columns = kwargs.pop('columns', ['default'])
        addnl_fields = kwargs.pop('addnl_fields', [])
        user_query = kwargs.pop('query_str', '')
        os = kwargs.pop('os', [])
        model = kwargs.pop('model', [])
        vendor = kwargs.pop('vendor', [])
        os_version = kwargs.pop('version', [])
        namespace = kwargs.pop('namespace', [])

        drop_cols = []

        if os or model or vendor or os_version:
            df = self._get_table_sqobj('device').get(
                columns=['namespace', 'hostname', 'os', 'model', 'vendor',
                         'version'],
                os=os,
                model=model,
                vendor=vendor,
                version=os_version,
                namespace=namespace,
                **kwargs)
        else:
            df = self._get_table_sqobj('device').get(
                columns=['namespace', 'hostname'], namespace=namespace,
                **kwargs)

        if df.empty:
            return pd.DataFrame()

        namespace = df.namespace.unique().tolist()
        hosts = df.hostname.unique().tolist()
        dev_nsgrp = df.groupby(['namespace'])

        # Get list of namespaces we're polling
        pollerdf = self._get_table_sqobj('sqPoller') \
            .get(columns=['namespace', 'hostname', 'service', 'status', 'timestamp'],
                 namespace=namespace, hostname=hosts)

        if pollerdf.empty:
            return pd.DataFrame()

        nsgrp = pollerdf.groupby(by=['namespace'])
        pollerns = sorted(pollerdf.namespace.unique().tolist())
        newdf = pd.DataFrame({
            'namespace': pollerns,
            'deviceCnt': dev_nsgrp['hostname'].nunique().tolist(),
            'serviceCnt': nsgrp['service'].nunique().tolist()
        })
        errsvc_df = pollerdf.query('status != 0 and status != 200') \
                            .groupby(by=['namespace'])['service'] \
                            .nunique().reset_index()
        newdf = newdf.merge(errsvc_df, on=['namespace'], how='left',
                            suffixes=['', '_y']) \
            .rename({'service': 'errSvcCnt'}, axis=1) \
            .fillna({'errSvcCnt': 0})
        newdf['errSvcCnt'] = newdf['errSvcCnt'].astype(int)

        # What protocols exist
        for table, fld in [('ospf', 'hasOspf'), ('bgp', 'hasBgp'),
                           ('evpnVni', 'hasVxlan'), ('mlag', 'hasMlag')]:
            df = self._get_table_sqobj(table) \
                .get(namespace=pollerns, hostname=hosts,
                     columns=['namespace'])
            if df.empty:
                newdf[fld] = False
                continue

            gotns = df.namespace.unique().tolist()
            newdf[fld] = newdf.apply(
                lambda x, y: True if x.namespace in y else False,
                axis=1, args=(gotns,))

        newdf = self._handle_user_query_str(newdf, user_query)

        # Look for the rest of info only in selected namepaces
        newdf['lastUpdate'] = nsgrp['timestamp'].max() \
            .reset_index()['timestamp']

        return newdf.drop(columns=drop_cols)

    def find(self, **kwargs):
        '''Find the information requsted:

        address: given a MAC or IP address, find the first hop switch its 
                 connected to
        '''

        addrlist = kwargs.pop('address', [])
        columns = kwargs.pop('columns', ['default'])
        query_str = kwargs.pop('query_str', '')

        dflist = []
        if isinstance(addrlist, str):
            addrlist = [addrlist]
        for addr in addrlist:
            df = self._find_address(addr, **kwargs)
            if not df.empty:
                dflist.append(df)
        if dflist:
            df = pd.concat(dflist)

        if not df.empty and query_str:
            return df.query(query_str)

        return df

    def summarize(self, **kwargs):
        '''Summarize for network

        Summarize for network is a bit different from the rest of the
        summaries because its not grouped by namespace.
        '''

        df = self.get(**kwargs)

        if df.empty:
            return df

        self.ns = pd.DataFrame({
            'namespacesCnt': [df.namespace.count()],
            'servicePerNsStat': [[df.serviceCnt.min(), df.serviceCnt.max(),
                                  df.serviceCnt.median()]],
            'nsWithMlagCnt': [df.loc[df.hasMlag].shape[0]],
            'nsWithBgpCnt': [df.loc[df.hasBgp].shape[0]],
            'nsWithOspfCnt': [df.loc[df.hasOspf].shape[0]],
            'nsWithVxlanCnt': [df.loc[df.hasVxlan].shape[0]],
            'nsWithErrsvcCnt': [df.loc[df.errSvcCnt != 0].shape[0]],
        })

        # At this point we have a single row with index 0, so rename
        return self.ns.T.rename(columns={0: 'summary'})

    def _find_asn(self, asn: List[str], **kwargs) -> pd.DataFrame:
        """Find the hosts with the ASN listed

        Args:
            asn (List[str]): List of ASNs, space separated

        Returns:
            pd.DataFrame: The dataframe with the info. The columns are-
                          namespace, hostname, asn
        """

        if not asn:
            return pd.DataFrame()

        _ = kwargs.pop('vlan', [])
        try:
            asn = [int(x) for x in asn]
        except ValueError:
            return pd.DataFrame({'error': ['ASNs must be an integers']})

        return self._get_table_sqobj('bgp') \
            .get(asn=asn, columns=['default'], **kwargs)

    def _find_address(self, addr: str, **kwargs) -> pd.DataFrame:
        """Find the origin or the first hop network device that owns this addr

        Args:
            addr (str): address to search for

        Returns:
            pd.DataFrame: Dataframe with the relevant information
        """

        cols = ['namespace', 'hostname', 'ifname',
                'vrf', 'ipAddress', 'vlan', 'macaddr', 'how', 'timestamp']

        vlan = kwargs.pop('vlan', '')
        vrf = kwargs.pop('vrf', '')

        # Convert Cisco-style MAC address to standard MAC addr format,
        # and lowercase all letters of the alphabet
        if '.' in addr:
            addr = convert_macaddr_format_to_colon(addr)
        addr = addr.lower()

        addr_df = self._find_addr_arp(addr, **kwargs)
        if addr_df.empty:
            return addr_df
        addr_df = self._find_first_hop_attach(addr_df)
        if vlan and not addr_df.empty:
            addr_df = addr_df.query(f'vlan == {vlan}')
        if vrf and not addr_df.empty:
            addr_df = addr_df.query(f'vrf == "{vrf}"')

        if addr_df.empty:
            return addr_df

        # If routed interface and unnumbered interfaces are used, we need to
        # identify the primary interface
        addr_df = addr_df \
            .dropna() \
            .drop_duplicates(subset=['namespace', 'hostname', 'vrf']) \
            .reset_index(drop=True)

        if addr_df.type.unique().tolist() == ['routed']:
            addr_df = self._find_primary_interface(addr_df, addr)

        return addr_df

    def _find_bond_members(self, namespace: str, hostname: str,
                           ifname: str, timestamp: int) -> List[str]:
        """This function returns the member ports of a bond interface
        Args:
            namespace (str): Namespace of the host
            hostname (str): Hostname of the host
            ifname (str): Interface name of the bond
            timestamp (int): Timestamp (epoch secs), at which to look for

        Returns:
            mbr_list(List[str]): List of member ports
        """

        # get list  of namespaces, hostnames and ifnames to get info for

        ifdf = self._get_table_sqobj('interfaces',
                                     start_time=str(timestamp-30),
                                     end_time=str(timestamp+30)) \
            .get(namespace=[namespace], hostname=[hostname], master=[ifname])

        if not ifdf.empty:
            return ifdf.ifname.unique().tolist()

        return [ifname]

    def _find_addr_arp(self, addr: str, **kwargs) -> pd.DataFrame:
        """Find the origin or the first hop network device that owns this addr

        Args:
            addr (str): address to search for

        Returns:
            pd.DataFrame: Dataframe with the relevant information
        """

        cols = ['namespace', 'hostname', 'ifname',
                'vrf', 'ipAddress', 'macaddr', 'timestamp']

        if any(x in addr for x in ['::', '.']):
            arpdf = self._get_table_sqobj('arpnd').get(
                ipAddress=addr, **kwargs)
        else:
            arpdf = self._get_table_sqobj('arpnd').get(
                macaddr=addr, **kwargs)

        result = []
        for row in arpdf.itertuples():
            if row.oif.endswith('-v0'):
                # Handle VRR interfaces in Cumulus
                oif = row.oif[:-3]
            else:
                oif = row.oif
            ifdf = self._get_table_sqobj(
                'interfaces',
                start_time=str(row.timestamp.timestamp()-30),
                end_time=str(row.timestamp.timestamp()+30)) \
                .get(namespace=[row.namespace], hostname=[row.hostname],
                     ifname=[oif])

            if not ifdf.empty:
                if 'vlan' not in ifdf.type.unique():
                    # Routed interface
                    result.append({
                        'namespace': row.namespace,
                        'hostname': row.hostname,
                        'vrf': ifdf.master.unique().tolist()[0] or 'default',
                        'ipAddress': row.ipAddress,
                        'vlan': ifdf.vlan.astype(str).unique().tolist()[0],
                        'macaddr': row.macaddr,
                        'ifname': oif,
                        'type': 'routed',
                        'timestamp': row.timestamp
                    })
                    continue

                macdf = self._get_table_sqobj(
                    'macs',
                    start_time=str(row.timestamp.timestamp()-30),
                    end_time=str(row.timestamp.timestamp()+30)) \
                    .get(namespace=[row.namespace], hostname=[row.hostname],
                         vlan=ifdf.vlan.astype(str).unique().tolist(),
                         macaddr=[row.macaddr],
                         columns=['default'],
                         localOnly=True)
                if not macdf.empty:
                    oifs = [x for x in macdf.oif.unique() if x !=
                            "vPC Peer-Link"]
                    if not oifs:
                        continue
                    result.append({
                        'namespace': row.namespace,
                        'hostname': row.hostname,
                        'vrf': ifdf.master.unique().tolist()[0] or 'default',
                        'ipAddress': row.ipAddress,
                        'vlan': ifdf.vlan.astype(str).unique().tolist()[0],
                        'macaddr': row.macaddr,
                        'ifname': oifs[0],
                        'type': 'bridged',
                        'timestamp': row.timestamp
                    })

        return pd.DataFrame(result)

    def _find_first_hop_attach(self, addr_df: pd.DataFrame) -> pd.DataFrame:
        """Find the first hop switch attachment point for this address

        Args:
            addr_df (pd.DataFrame): Dataframe with the address information

        Returns:
            pd.DataFrame: Dataframe with the L2 attachment point information
        """

        # We'll need to resolve bond interfaces first
        l2_addr_df = addr_df
        if l2_addr_df.empty:
            return addr_df

        result = []
        for row in l2_addr_df.itertuples():
            do_continue = True
            match_ifname = row.ifname.split('.')[0]
            match_hostname = row.hostname
            match_namespace = row.namespace
            match_vrf = row.vrf
            match_endtime = str(row.timestamp.timestamp() + 30)
            match_starttime = str(row.timestamp.timestamp() - 30)
            macobj = self._get_table_sqobj('macs', start_time=match_starttime,
                                           end_time=match_endtime)
            lldpobj = self._get_table_sqobj('lldp', start_time=match_starttime,
                                            end_time=match_endtime)

            while True:
                # Do not attempt to replace tmp_df with l2_addr_df and get
                # rid of do_contiue. For some weird reason, with pandas ver
                # 1.2.5, the explode line below hangs if we just use
                # l2_addr_df

                mbr_ports = self._find_bond_members(
                    match_namespace, match_hostname, match_ifname,
                    row.timestamp.timestamp())

                lldp_df = lldpobj.get(namespace=[row.namespace],
                                      hostname=[row.hostname], ifname=mbr_ports)

                if not lldp_df.empty:
                    peer_host = lldp_df.peerHostname.unique().tolist()[0]
                    if row.type == 'routed':
                        match_hostname = peer_host
                        match_ifname = lldp_df.peerIfname.unique().tolist()[0]
                        mbr_ports = self._find_bond_members(
                            match_namespace, match_hostname, match_ifname,
                            row.timestamp.timestamp())
                        # Need to get VRF for the interface
                        ifdf = self._get_table_sqobj('interfaces') \
                            .get(namespace=[match_namespace],
                                 hostname=[match_hostname], ifname=[match_ifname])
                        if not ifdf.empty:
                            match_vrf = ifdf.master.unique().tolist()[
                                0] or 'default'
                        # Routed ports never go more than one hop
                        break

                    macdf = macobj.get(namespace=[match_namespace],
                                       hostname=[peer_host],
                                       macaddr=row.macaddr,
                                       vlan=str(row.vlan),
                                       columns=['namespace', 'hostname', 'vlan',
                                                'macaddr', 'oif'])
                    if not macdf.empty:
                        match_hostname = peer_host
                        match_ifname = macdf.oif.unique().tolist()[0]
                    else:
                        break
                else:
                    break

            if match_ifname in mbr_ports:
                mbr_ports = ''
            result.append({
                'namespace': row.namespace,
                'hostname': match_hostname,
                'vrf': match_vrf,
                'ipAddress': row.ipAddress,
                'vlan': row.vlan,
                'macaddr': row.macaddr,
                'ifname': match_ifname,
                'bondMembers': ', '.join(mbr_ports),
                'type': row.type,
                'timestamp': row.timestamp
            })

        if result:
            return pd.DataFrame(result)

        return addr_df

    def _find_primary_interface(self, addr_df: pd.DataFrame,
                                addr: str) -> pd.DataFrame:
        """Find the primary interface for this address

        Args:
            addr_df (pd.DataFrame): Dataframe with the address information
            addr (str): address to search for

        Returns:
            pd.DataFrame: Dataframe with primary interface rows only
        """

        hostnsgrp = addr_df.groupby(['hostname', 'namespace', 'vrf'])
        if len(hostnsgrp) > 1:
            # not a set of duplicated interfaces, return
            return addr_df

        # We have a set of duplicated interfaces, find the primary
        df = self._get_table_sqobj('interfaces') \
            .get(namespace=[addr_df.namespace[0]],
                 hostname=[addr_df.hostname[0]],
                 vrf=[addr_df.vrf[0]], type='loopback')
        if df.empty:
            return addr_df

        df = df.explode('ipAddressList').explode('ip6AddressList') \
            .query(f'ipAddressList.str.startswith("{addr}/") or '
                   f'ip6AddressList.str.startswith("{addr}/")') \
            .reset_index(drop=True)

        if not df.empty:
            addr_df['ifname'] = df.ifname.unique().tolist()[0]
            if df.type[0] == "bond":
                addr_df['bondMbrs'] = self._find_bond_members(
                    addr_df.namespace[0], addr_df.hostname[0],
                    addr_df.ifname[0], addr_df.timestamp[0].timestamp())
            else:
                addr_df['bondMbrs'] = ''

            return addr_df

        return addr_df



#!/usr/bin/env python

import os
try:
    import simplejson as json
except ModuleNotFoundError:
    import json
import re
import time
import argparse
from pathlib import Path
from netconan.ip_anonymization import IpAnonymizer, IpV6Anonymizer
from netconan.ip_anonymization import anonymize_ip_addr
from collections import defaultdict
from faker import Factory
from jsonpath_ng import parse


class SqAnonymizer(object):
    """Suzieq Gather-Once output Anonymizer Class

    This class implements the anonymizer for the Suzieq run-once=gather output
    format. This anonymizes:
    - IP adddresses (v4 and v6)
    - MAC addresses (in NXOS/EOS or standard format)
    - Hostnames (FQDN or otherwise)

    The workflow is to point either a single file to anonymize or a directory
    of files to anonymize. If its a directory, it only attempts to anonymize
    the files with .output filename and excludes those ending with _anon,output
    For every file in the directory, it produces the equivalent _anon.output
    file that contains the anonymized data. For example, lldp.output is turned
    into lldp_anon.output, device.output is turned into device_anon.output and
    so on. The _anon.output file is in the run-once=gather format.

    Since hostnames can appear anywhere, and do not have a unique format, we
    require users to provide additional information. This information consists
    of a JSONpath expression to the key that holds hostname or a prefix string
    that comes before every hostname. The former is required for JSON outputs &
    the latter for non-JSON output. This model is followed for all data that
    has a non-unique format such as ASNs, VRF names and so on. This is also
    true of MAC addresses in the EOS/NXOS format.

    To ensure the same output when run at different times, the user must
    provide a seed that remains the same across different runs. Otherwise, we
    pick the time of day as the seed which produces different outputs when run
    at diff times.
    """

    def __init__(self, seed: int = 0, preserve_v4_prefix: list = []):
        """Initiaize the anonymizer with a seed"""

        if not seed:
            seed = int(time.time())

        preserve_v4_prefix.append('0.0.0.0/0')
        self.fakeit = Factory.create()
        self.fakeit.seed(seed)
        salt = self.fakeit.password(length=16)
        self.anonv4 = IpAnonymizer(salt, preserve_prefixes=preserve_v4_prefix)
        self.anonv6 = IpV6Anonymizer(salt)
        self.hostmap = {}
        self.macmap = defaultdict(self.fakeit.mac_address)

        # Courtesy of Stackoverflow
        # (https://stackoverflow.com/questions/1418423/the-hostname-regex)
        self.hname_re = r'([a-zA-Z0-9](?:(?:[_a-zA-Z0-9-]*|(?<!-)\.(?![-.]))*[_a-zA-Z0-9]+)?)'

        # This regex will not catch the macaddr in this format:
        # "This is a macaddr:00:01:02:FF:cf:9b"
        # because the macaddr is preceded by :. It will also not match:
        # "00:01:29:AB:cf:47:01" because there's one more group (:01) after
        # the mac address
        self.macaddr_re = r'(?<!:)([0-9a-f]{2}(?::[0-9a-f]{2}){5})(?!:)'

        # NXOS and EOS use this format
        self.nxos_macaddr_re = r'([0-9A-F]{4}\.[0-9A-F]{4}\.[0-9A-F]{4})'

        self.preset_hostname_jpaths = [
            ['TABLE_nbor.ROW_nbor[*].chassis_id'],  # NXOS LLDP
            # NXOS CDP
            ['TABLE_cdp_neighbor_brief_info.ROW_cdp_neighbor_brief_info[*].device_id'],
            ['lldpNeighbors.*.lldpNeighborInfo[*].systemName'],  # EOS
            ['lldp[*].interface[*].chassis[*].name[*].value'],  # lldpd
            ['lldp-neighbors-information.[0].lldp-neighbor-information[*].lldp-remote-system-name.[0].data'],  # JunOS
            ['*.*.neighborCapabilities[*].hostName.advHostName'],  # FRR BGP
            ['*.*.neighborCapabilities[*].hostName.rcvHostName'],  # FRR BGP
            ['*.*.neighborCapabilities[*].hostName.advDomainName'],  # FRR BGP
            ['*.*.neighborCapabilities[*].hostName.rcvDomainName'],  # FRR BGP
            ['host_name']          # NXOS show version output
        ]

        self.preset_hostname_pfxlst = [
            ['SysName:'],
            ['Static hostname:'],
        ]

        # Needed for NXOS and EOS which use the non-standard MACADDR format
        self.preset_macaddr_jpaths = [
            ['configMacAddress'],  # EOS show version output
            ['systemMacAddress'],  # EOS show version output
            ['hwMacAddress'],      # EOS show version output
        ]

    def anonymize(self, file_or_dir: str, anonymize_ip=True,
                  anonymize_mac=True, anonymize_hostname=True,
                  host_prefixes: list = [], host_jpath: list = [],
                  mac_jpath: list = []):
        """Anonymizes the parameters desired in the given input file/dir """

        self.host_jpath = (host_jpath or []) + self.preset_hostname_jpaths
        self.host_pfxlst = (host_prefixes or []) + self.preset_hostname_pfxlst
        self.mac_jpath = (mac_jpath or []) + self.preset_macaddr_jpaths

        if os.path.isfile(file_or_dir):
            self.anonymize_file(file_or_dir, host_prefixes,
                                host_jpath, mac_jpath)
        else:
            # Invoke the command for each file in the directory
            for file in Path(file_or_dir).glob('*.output'):
                if str(file).endswith('_anon.output'):
                    continue
                self.anonymize_file(file, host_prefixes, host_jpath, mac_jpath)

    def anonymize_file(self, file: str, host_pfxlst: list,
                       host_jpath: list, mac_jpath, anonymize_ip=True,
                       anonymize_mac=True, anonymize_hostname=True):
        """Anonymize an individual file"""

        name, suffix = os.path.basename(file).split('.')
        dirname = os.path.dirname(file)
        anon_fname = f'{dirname}/{name}_anon.{suffix}'

        with open(file, 'r') as f:
            data = f.read()

        # Prepare output file by truncating it
        with open(anon_fname, 'w') as f:
            f.write('')

        # The run-once= gather output is not proper JSON. Its
        # written out as a list of entries, but there's no
        # overarching list covering this list. So, we have to
        # handcraft the entries to get them to be readable as
        # JSON. The first entry is missing an ending ']' and the
        # last entry is missing an opening '[' while the intermediate
        # entries are missing both the starting and ending '['.
        entries = re.split(r'\]\n*\[', data)
        entlen = len(entries)
        new_entries = []

        for i, elem in enumerate(entries):
            newelem = elem.replace('\n', '').strip()
            try:
                jelem = json.loads(newelem)
                new_entries.append(jelem)
            except Exception as e:
                if i == 0:
                    newelem = newelem + ']'
                elif i == entlen - 1:
                    newelem = '[' + newelem
                else:
                    newelem = '[' + newelem + ']'

            try:
                jelem = json.loads(newelem)
            except Exception as e:
                print(f"JSON load of {i} item failed with error {str(e)}")
                jelem = []

            for item in jelem:
                # Populate the host name anonymizer as much as possible
                item['hostname'] = re.sub(self.hname_re, self.get_anonhost,
                                          item['hostname'])
        for jelem in new_entries:

            for item in jelem:
                if isinstance(item['data'], str):
                    try:
                        jdata = json.loads(item['data'])
                        is_json = True
                    except json.decoder.JSONDecodeError:
                        # non-JSON command output
                        jdata = item['data']
                        is_json = False
                elif isinstance(item['data'], dict):
                    # EOS LLDP output looks like this
                    jdata = item['data']
                    is_json = True

                anondata = ''
                if jdata:
                    if is_json:
                        anondata = json.dumps(jdata)
                    else:
                        anondata = jdata

                if anonymize_mac:
                    anondata = self.anonymize_mac(
                        self.macmap, anondata, is_json)

                if anonymize_ip:
                    anondata = self.anonymize_ip(item, anondata)

                if anonymize_hostname:
                    anondata = self.anonymize_hostname(item, anondata,
                                                       is_json)

                if anondata:
                    item['data'] = anondata

            # Append each entry
            with open(anon_fname, 'a') as f:
                f.write(json.dumps(jelem, indent=4))

    def anonymize_ip(self, item: dict, anondata: str) -> str:
        """Anonymize the IP address in the data and in the item header

        Anonymize the IP addresses, v4 and v6, in the supplied data. Also
        anonymize the IP address in the data header. Both are specific to the
        run-once=gather format of Suzieq
        """
        item['address'] = anonymize_ip_addr(self.anonv4, item['address'])
        item['address'] = anonymize_ip_addr(self.anonv6, item['address'])

        if not anondata:
            return anondata

        anondata = anonymize_ip_addr(self.anonv4, anondata, False)
        anondata = anonymize_ip_addr(self.anonv6, anondata, False)

        return anondata

    def anonymize_mac(self, macmap, anondata: str, is_json: bool) -> str:
        """Anonymize the MAC address in the data and in the item header

        Both are specific to the run-once=gather format of Suzieq
        """
        def nxos_mac_sub(match):
            newmac = macmap[match.group(1)]
            newmac = newmac.split(':')
            newmac = (''.join(newmac[0:2]) + '.' + ''.join(newmac[2:4]) + '.' +
                      ''.join(newmac[4:]))
            return newmac

        def jp_anon_macaddr(orig_v, orig_kv, orig_k):
            """JSONPATH update macaddress in NXOS/EOS format"""
            orig_kv[orig_k] = macmap[orig_v]

        if not anondata:
            return

        anondata = re.sub(self.macaddr_re,
                          lambda x: macmap[x.group(1)],
                          anondata, flags=re.IGNORECASE)

        anondata = re.sub(self.nxos_macaddr_re, nxos_mac_sub, anondata,
                          flags=re.IGNORECASE)

        if is_json:
            janon = json.loads(anondata)
            for path in self.mac_jpath:
                jp = parse(path[0])
                jp.update(janon, jp_anon_macaddr)
            anondata = json.dumps(janon)

        return anondata

    def get_anonhost(self, host: str) -> str:
        """Check various conditions to return the anonymized hostname.

        Sometimes hostnames are present without FQDN and sometimes with.
        We want to map them consistently to the same anonymized hostname.
        We won't know which format we'll encounter first and so we have to
        handle that as well. Hence this routine
        """

        if not host:
            return host

        if isinstance(host, re.Match):
            orighost = host.group(1)
        else:
            orighost = host

        helem = orighost.split('.')
        anonhost = self.hostmap.get(orighost, '')
        if not anonhost and len(helem) > 1:
            anonhost = self.hostmap.get(helem[0], '')

        if not anonhost:
            anonhost = self.fakeit.hostname()
            if len(helem) > 1:
                self.hostmap[orighost] = anonhost
                self.hostmap[helem[0]] = anonhost
            else:
                self.hostmap[orighost] = anonhost

        if len(helem) == 1:
            return anonhost.split('.')[0]
        else:
            return anonhost

    def anonymize_hostname(self, item: dict, anondata: str, is_json: bool) \
            -> str:
        """Anonymize the hostname in the item header and the data

        """

        def anon_hostname(match):
            newhostname = self.get_anonhost(match.group(1))
            if is_json:
                return f'"{kwd[0]}": "{newhostname}"'
            else:
                return f'{kwd[0]} {newhostname}'

        def jp_anon_hostname(orig_v, orig_kv, orig_k):
            """JSONPATH update hostname"""
            if orig_v == 'n/a':  # handle null domainname in FRR BGP
                return
            orig_kv[orig_k] = self.get_anonhost(orig_v)

        if not anondata:
            return anondata

        if is_json:
            janon = json.loads(anondata)
            for path in self.host_jpath:
                jp = parse(path[0])
                jp.update(janon, jp_anon_hostname)
            anondata = json.dumps(janon)
        else:
            for kwd in self.host_pfxlst:
                hpat = r'{}\s+{}'.format(kwd[0], self.hname_re)
                anondata = re.sub(hpat, anon_hostname, anondata)

        # The hostname shows up in strange places and so do one more pass with
        # their hostnames we have to see if we can catch those entries
        for hostname in self.hostmap:
            anondata = anondata.replace(hostname, self.hostmap[hostname])
        return anondata

    def dump_mappings(self, file_or_dir) -> None:
        """Dump the mappings of various fields into a file"""

        if os.path.isdir(file_or_dir):
            dirname = file_or_dir
        else:
            dirname = os.path.dirname(os.path.abspath(file_or_dir))

        with open(f'{dirname}/mapping.txt', 'w') as f:
            for k, v in self.hostmap.items():
                f.write(f'{k}\t{v}\n')
            for k, v in self.macmap.items():
                f.write(f'{k}\t{v}\n')
            self.anonv4.dump_to_file(f)
            self.anonv6.dump_to_file(f)


def anonymizer_main():

    parser = argparse.ArgumentParser(
        description="Anonymize suzieq run-once data files")
    parser.add_argument(
        "-i",
        "--input",
        type=str, required=True,
        help="Directory of files or single File to anonymize",
    )
    parser.add_argument(
        "-s",
        "--seed",
        type=str,
        help="Value to seed the randomizer with"
    )
    parser.add_argument(
        "-ph",
        "--host-prefix",
        nargs="+", action='append',
        help="Keywords preceding hostname to anonymize hostname (non-JSON output)",
    )
    parser.add_argument(
        "-PH",
        "--host-jpath",
        nargs="+", action='append',
        help=("JSON path to get to hostname in JSON output (use option "
              "multiple times to provide multiple values)")
    )
    parser.add_argument(
        "-PM",
        "--mac-jpath",
        nargs="+", action='append',
        help=("JSON path to get to NXOS/EOS Macaddr in JSON output"
              "Use option multiple times to provide multiple values")
    )
    parser.add_argument(
        "-ip",
        "--preserve-v4-prefixes",
        nargs="+", action='append',
        help=("IP address prefix to NOT anonymize (default is always incl)"
              "Use option multiple times to provide multiple values")
    )
    userargs = parser.parse_args()

    anonymizer = SqAnonymizer(userargs.seed)
    anonymizer.anonymize(userargs.input, host_prefixes=userargs.host_prefix,
                         host_jpath=userargs.host_jpath,
                         mac_jpath=userargs.mac_jpath)
    anonymizer.dump_mappings(userargs.input)


if __name__ == '__main__':
    anonymizer_main()



import time
import ipaddress
from nubia import command, argument
import pandas as pd

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.routes import RoutesObj


@command("route", help="Act on Routes", aliases=['routes'])
class RouteCmd(SqCommand):
    """Routing table information"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        query_str: str = ' ',
        columns: str = "default",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=RoutesObj,
        )

    def _json_print_handler(self, input):
        """This handler calls the code to print the IPNetwork as a string"""
        if isinstance(input, ipaddress.IPv4Network):
            return ipaddress.IPv4Network.__str__(input)
        elif isinstance(input, ipaddress.IPv6Network):
            return ipaddress.IPv6Network.__str__(input)
        return input

    def _get_ipvers(self, value: str) -> int:
        """Return the IP version in use"""

        if ':' in value:
            ipvers = 6
        elif '.' in value:
            ipvers = 4
        else:
            ipvers = ''

        return ipvers

    @command("show")
    @argument("prefix", description="Prefix, in quotes, to filter show on")
    @argument("vrf", description="VRF to qualify")
    @argument("protocol", description="routing protocol to qualify")
    @argument("prefixlen", description="must be of the form "
              "[<|<=|>=|>|!] length")
    def show(self, prefix: str = "", vrf: str = '', protocol: str = "",
             prefixlen: str = ""):
        """Show Routing table info
        """
        if self.columns is None:
            return

        # Get the default display field names
        now = time.time()

        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname,
                                prefix=prefix.split(),
                                vrf=vrf.split(),
                                protocol=protocol.split(),
                                columns=self.columns,
                                namespace=self.namespace,
                                prefixlen=prefixlen,
                                query_str=self.query_str,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)

    @command("summarize")
    @argument("vrf", description="VRF to qualify")
    def summarize(self, vrf: str = ''):
        """Summarize key routing table info
        """
        # Get the default display field names
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.summarize,
                                hostname=self.hostname,
                                vrf=vrf.split(),
                                namespace=self.namespace,
                                query_str=self.query_str,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df, json_orient='columns')

    @command('lpm')
    @argument("address", description="IP Address, in quotes, for lpm query")
    @argument("vrf", description="specific VRF to qualify")
    def lpm(self, address: str = '', vrf: str = ''):
        """Show the Longest Prefix Match(LPM) on a given prefix, vrf
        """
        if self.columns is None:
            return

        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        if not address:
            print('address is mandatory parameter')
            return

        df = self._invoke_sqobj(self.sqobj.lpm,
                                hostname=self.hostname,
                                address=address,
                                vrf=vrf.split(),
                                ipvers=self._get_ipvers(address),
                                columns=self.columns,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)



import os
import re
import sys
from typing import Tuple, List
from pathlib import Path
import logging
from logging.handlers import RotatingFileHandler
import json
import yaml
from dateutil.relativedelta import relativedelta
from datetime import datetime
from tzlocal import get_localzone
from pytz import all_timezones
import fcntl
from importlib.util import find_spec
import errno
from dateparser import parse
from itertools import groupby

import pandas as pd
import pyarrow as pa


logger = logging.getLogger(__name__)


MAX_MTU = 9216


def validate_sq_config(cfg):
    """Validate Suzieq config file

    Parameters:
    -----------
    cfg: yaml object, YAML encoding of the config file

    Returns:
    --------
    status: None if all is good or error string
    """

    if not isinstance(cfg, dict):
        return "FATAL: Invalid config file format"

    ddir = cfg.get("data-directory", None)
    if not ddir:
        return "FATAL: No data directory for output files specified"

    if not os.path.isdir(ddir):
        os.makedirs(ddir, exist_ok=True)

    if (not os.path.isdir(ddir) or not (os.access(ddir, os.R_OK | os.W_OK |
                                                  os.EX_OK))):
        return f'FATAL: Data directory {ddir} is not an acceesible dir'

    # Locate the service and schema directories
    svcdir = cfg.get('service-directory', None)
    if (not (svcdir and os.path.isdir(ddir) and
             os.access(svcdir, os.R_OK | os.W_OK | os.EX_OK))):
        sqdir = get_sq_install_dir()
        svcdir = f'{sqdir}/config'
        if os.access(svcdir, os.R_OK | os.EX_OK):
            cfg['service-directory'] = svcdir
        else:
            svcdir = None

    if not svcdir:
        return 'FATAL: No service directory found'

    schemadir = cfg.get('schema-directory', None)
    if not (schemadir and os.access(schemadir, os.R_OK | os.EX_OK)):
        schemadir = f'{svcdir}/schema'
        if os.access(schemadir, os.R_OK | os.EX_OK):
            cfg['schema-directory'] = schemadir
        else:
            schemadir = None

    if not schemadir:
        return 'FATAL: No schema directory found'

    # Move older format logging level and period to appropriate new location
    if 'poller' not in cfg:
        cfg['poller'] = {}

    for knob in ['logging-level', 'period']:
        if knob in cfg:
            cfg['poller'][knob] = cfg[knob]

    if 'rest' not in cfg:
        cfg['rest'] = {}

    for knob in ['API_KEY', 'rest_certfile', 'rest_keyfile']:
        if knob in cfg:
            cfg['rest'][knob] = cfg[knob]

    # Verify timezone if present is valid
    def_tz = get_localzone().zone
    reader = cfg.get('analyzer', {})
    if reader and isinstance(reader, dict):
        usertz = reader.get('timezone', '')
        if usertz and usertz not in all_timezones:
            return f'Invalid timezone: {usertz}'
        elif not usertz:
            reader['timezone'] = def_tz
    else:
        cfg['analyzer'] = {'timezone': def_tz}

    return None


def load_sq_config(validate=True, config_file=None):
    """Load (and validate) basic suzieq config"""

    # Order of looking up suzieq config:
    #   Current directory
    #   ${HOME}/.suzieq/

    cfgfile = None
    cfg = None

    cfgfile = sq_get_config_file(config_file)

    if cfgfile:
        try:
            with open(cfgfile, "r") as f:
                cfg = yaml.safe_load(f.read())
        except Exception as e:
            print(f'ERROR: Unable to open config file {cfgfile}: {e.args[1]}')
            sys.exit(1)

        if not cfg:
            print(f'ERROR: Empty config file {cfgfile}')
            sys.exit(1)

        if validate:
            error_str = validate_sq_config(cfg)
            if error_str:
                print(f'ERROR: Invalid config file: {config_file}')
                print(error_str)
                sys.exit(1)

    if not cfg:
        print(f"suzieq requires a configuration file either in ./suzieq-cfg.yml "
              "or ~/suzieq/suzieq-cfg.yml")
        sys.exit(1)

    return cfg


def sq_get_config_file(config_file):
    """Get the path to the suzieq config file"""

    if config_file:
        cfgfile = config_file
    elif os.path.exists("./suzieq-cfg.yml"):
        cfgfile = "./suzieq-cfg.yml"
    elif os.path.exists(os.getenv("HOME") + "/.suzieq/suzieq-cfg.yml"):
        cfgfile = os.getenv("HOME") + "/.suzieq/suzieq-cfg.yml"
    else:
        cfgfile = None
    return cfgfile


class Schema(object):
    def __init__(self, schema_dir):
        self._init_schemas(schema_dir)
        if not self._schema:
            raise ValueError("Unable to get schemas")

    def _init_schemas(self, schema_dir: str):
        """Returns the schema definition which is the fields of a table"""

        schemas = {}
        phy_tables = {}
        types = {}

        if not (schema_dir and os.path.exists(schema_dir)):
            logger.error(
                "Schema directory {} does not exist".format(schema_dir))
            raise Exception(f"Schema directory {schema_dir} does not exist")

        for root, _, files in os.walk(schema_dir):
            for topic in files:
                if not topic.endswith(".avsc"):
                    continue
                with open(root + "/" + topic, "r") as f:
                    data = json.loads(f.read())
                    table = data["name"]
                    schemas[table] = data["fields"]
                    types[table] = data.get('recordType', 'record')
                    phy_tables[data["name"]] = data.get("physicalTable", table)
            break

        self._schema = schemas
        self._phy_tables = phy_tables
        self._types = types

    def tables(self):
        return self._schema.keys()

    def fields_for_table(self, table):
        return [f['name'] for f in self._schema[table]]

    def get_raw_schema(self, table):
        return self._schema[table]

    def field_for_table(self, table, field):
        for f in self._schema[table]:
            if f['name'] == field:
                return f

    def type_for_table(self, table):
        return self._types[table]

    def key_fields_for_table(self, table):
        # return [f['name'] for f in self._schema[table] if f.get('key', None) is not None]
        return self._sort_fields_for_table(table, 'key')

    def augmented_fields_for_table(self, table):
        return [x['name'] for x in self._schema.get(table, [])
                if 'depends' in x]

    def sorted_display_fields_for_table(self, table, getall=False):
        return self._sort_fields_for_table(table, 'display', getall)

    def _sort_fields_for_table(self, table, tag, getall=False):
        fields = self.fields_for_table(table)
        field_weights = {}
        for f_name in fields:
            field = self.field_for_table(table, f_name)
            if field.get(tag, None) is not None:
                field_weights[f_name] = field.get(tag, 1000)
            elif getall:
                field_weights[f_name] = 1000
        return [k for k in sorted(field_weights.keys(),
                                  key=lambda x: field_weights[x])]

    def array_fields_for_table(self, table):
        fields = self.fields_for_table(table)
        arrays = []
        for f_name in fields:
            field = self.field_for_table(table, f_name)
            if isinstance(field['type'], dict) and field['type'].get('type', None) == 'array':
                arrays.append(f_name)
        return arrays

    def get_phy_table_for_table(self, table):
        """Return the name of the underlying physical table"""
        if self._phy_tables:
            return self._phy_tables.get(table, table)

    def get_partition_columns_for_table(self, table):
        """Return the list of partition columns for table"""
        if self._phy_tables:
            return self._sort_fields_for_table(table, 'partition')

    def get_arrow_schema(self, table):
        """Convert the internal AVRO schema into the equivalent PyArrow schema"""

        avro_sch = self._schema.get(table, None)
        if not avro_sch:
            raise AttributeError(f"No schema found for {table}")

        arsc_fields = []

        map_type = {
            "string": pa.string(),
            "long": pa.int64(),
            "int": pa.int32(),
            "double": pa.float64(),
            "float": pa.float32(),
            "timestamp": pa.int64(),
            "timedelta64[s]": pa.float64(),
            "boolean": pa.bool_(),
            "array.string": pa.list_(pa.string()),
            "array.nexthopList": pa.list_(pa.struct([('nexthop', pa.string()),
                                                     ('oif', pa.string()),
                                                     ('weight', pa.int32())])),
            "array.long": pa.list_(pa.int64()),
            "array.float": pa.list_(pa.float32()),
        }

        for fld in avro_sch:
            if "depends" in fld:
                # These are augmented fields, not in arrow"
                continue
            if isinstance(fld["type"], dict):
                if fld["type"]["type"] == "array":
                    if fld["type"]["items"]["type"] == "record":
                        avtype: str = "array.{}".format(fld["name"])
                    else:
                        avtype: str = "array.{}".format(
                            fld["type"]["items"]["type"])
                else:
                    # We don't support map yet
                    raise AttributeError
            else:
                avtype: str = fld["type"]

            arsc_fields.append(pa.field(fld["name"], map_type[avtype]))

        return pa.schema(arsc_fields)

    def get_parent_fields(self, table, field):
        avro_sch = self._schema.get(table, None)
        if not avro_sch:
            raise AttributeError(f"No schema found for {table}")

        for fld in avro_sch:
            if fld['name'] == field:
                if "depends" in fld:
                    return fld['depends'].split()
                else:
                    return []
        return []


class SchemaForTable(object):
    def __init__(self, table, schema: Schema = None, schema_dir=None):
        if schema:
            if isinstance(schema, Schema):
                self._all_schemas = schema
            else:
                raise ValueError("Passing non-Schema type for schema")
        else:
            self._all_schemas = Schema(schema_dir=schema_dir)
        self._table = table
        if table not in self._all_schemas.tables():
            raise ValueError(f"Unknown table {table}, no schema found for it")

    @property
    def type(self):
        return self._all_schemas.type_for_table(self._table)

    @property
    def version(self):
        return self._all_schemas.field_for_table(self._table,
                                                 'sqvers')['default']

    @property
    def fields(self):
        return self._all_schemas.fields_for_table(self._table)

    def get_phy_table(self):
        return self._all_schemas.get_phy_table_for_table(self._table)

    def get_partition_columns(self):
        return self._all_schemas.get_partition_columns_for_table(self._table)

    def key_fields(self):
        return self._all_schemas.key_fields_for_table(self._table)

    def get_augmented_fields(self):
        return self._all_schemas.augmented_fields_for_table(self._table)

    def sorted_display_fields(self, getall=False):
        return self._all_schemas.sorted_display_fields_for_table(self._table,
                                                                 getall)

    @property
    def array_fields(self):
        return self._all_schemas.array_fields_for_table(self._table)

    def field(self, field):
        return self._all_schemas.field_for_table(self._table, field)

    def get_display_fields(self, columns: list) -> list:
        """Return the list of display fields for the given table"""
        if columns == ["default"]:
            fields = self.sorted_display_fields()

            if "namespace" not in fields:
                fields.insert(0, "namespace")
        elif columns == ["*"]:
            fields = self.sorted_display_fields(getall=True)
        else:
            fields = [f for f in columns if f in self.fields]

        return fields

    def get_phy_table_for_table(self) -> str:
        """Return the underlying physical table for this logical table"""
        return self._all_schemas.get_phy_table_for_table(self._table)

    def get_raw_schema(self):
        return self._all_schemas.get_raw_schema(self._table)

    def get_arrow_schema(self):
        return self._all_schemas.get_arrow_schema(self._table)

    def get_parent_fields(self, field):
        return self._all_schemas.get_parent_fields(self._table, field)


def get_latest_files(folder, start="", end="", view="latest") -> list:
    lsd = []

    if start:
        ssecs = pd.to_datetime(
            start, infer_datetime_format=True).timestamp() * 1000
    else:
        ssecs = 0

    if end:
        esecs = pd.to_datetime(
            end, infer_datetime_format=True).timestamp() * 1000
    else:
        esecs = 0

    ts_dirs = False
    pq_files = False

    for root, dirs, files in os.walk(folder):
        flst = None
        if dirs and dirs[0].startswith("timestamp") and not pq_files:
            flst = get_latest_ts_dirs(dirs, ssecs, esecs, view)
            ts_dirs = True
        elif files and not ts_dirs:
            flst = get_latest_pq_files(files, root, ssecs, esecs, view)
            pq_files = True

        if flst:
            lsd.append(os.path.join(root, flst[-1]))

    return lsd


def get_latest_ts_dirs(dirs, ssecs, esecs, view):
    newdirs = None

    if not ssecs and not esecs:
        dirs.sort(key=lambda x: int(x.split("=")[1]))
        newdirs = dirs
    elif ssecs and not esecs:
        newdirs = list(filter(lambda x: int(x.split("=")[1]) > ssecs, dirs))
        if not newdirs and view != "changes":
            # FInd the entry most adjacent to this one
            newdirs = list(filter(lambda x: int(
                x.split("=")[1]) < ssecs, dirs))
    elif esecs and not ssecs:
        newdirs = list(filter(lambda x: int(x.split("=")[1]) < esecs, dirs))
    else:
        newdirs = list(
            filter(
                lambda x: int(x.split("=")[1]) < esecs and int(
                    x.split("=")[1]) > ssecs,
                dirs,
            )
        )
        if not newdirs and view != "changes":
            # FInd the entry most adjacent to this one
            newdirs = list(filter(lambda x: int(
                x.split("=")[1]) < ssecs, dirs))

    return newdirs


def get_latest_pq_files(files, root, ssecs, esecs, view):

    newfiles = None

    if not ssecs and not esecs:
        files.sort(key=lambda x: os.path.getctime("%s/%s" % (root, x)))
        newfiles = files
    elif ssecs and not esecs:
        newfiles = list(
            filter(lambda x: os.path.getctime(
                "%s/%s" % (root, x)) > ssecs, files)
        )
        if not newfiles and view != "changes":
            # FInd the entry most adjacent to this one
            newfiles = list(
                filter(
                    lambda x: os.path.getctime(
                        "{}/{}".format(root, x)) < ssecs, files
                )
            )
    elif esecs and not ssecs:
        newfiles = list(
            filter(lambda x: os.path.getctime(
                "%s/%s" % (root, x)) < esecs, files)
        )
    else:
        newfiles = list(
            filter(
                lambda x: os.path.getctime("%s/%s" % (root, x)) < esecs
                and os.path.getctime("%s/%s" % (root, x)) > ssecs,
                files,
            )
        )
        if not newfiles and view != "changes":
            # Find the entry most adjacent to this one
            newfiles = list(
                filter(lambda x: os.path.getctime(
                    "%s/%s" % (root, x)) < ssecs, files)
            )
    return newfiles


def calc_avg(oldval, newval):
    if not oldval:
        return newval

    return float((oldval+newval)/2)


def get_timestamp_from_cisco_time(input, timestamp):
    """Get timestamp in ms from the Cisco-specific timestamp string
    Examples of Cisco timestamp str are P2DT14H45M16S, P1M17DT4H49M50S etc.
    """
    if not input.startswith('P'):
        return 0
    months = days = hours = mins = secs = 0

    if 'T' in input:
        day, timestr = input[1:].split('T')
    else:
        day = input[1:]
        timestr = ''

    if 'Y' in day:
        years, day = day.split('Y')
        months = int(years)*12

    if 'M' in day:
        mnt, day = day.split('M')
        months = months + int(mnt)
    if 'D' in day:
        days = int(day.split('D')[0])

    if 'H' in timestr:
        hours, timestr = timestr.split('H')
        hours = int(hours)
    if 'M' in timestr:
        mins, timestr = timestr.split('M')
        mins = int(mins)
    if 'S' in timestr:
        secs = timestr.split('S')[0]
        secs = int(secs)

    delta = relativedelta(months=months, days=days,
                          hours=hours, minutes=mins, seconds=secs)
    return int((datetime.fromtimestamp(timestamp)-delta).timestamp()*1000)


def get_timestamp_from_junos_time(input, timestamp: int):
    """Get timestamp in ms from the Junos-specific timestamp string
    The expected input looks like: "attributes" : {"junos:seconds" : "0"}.
    We don't check for format because we're assuming the input would be blank
    if it wasn't the right format. The input can either be a dictionary or a
    JSON string.
    """

    if not input:
        # Happens for logical interfaces such as gr-0/0/0
        secs = 0
    else:
        try:
            if isinstance(input, str):
                data = json.loads(input)
            else:
                data = input
            secs = int(data.get('junos:seconds', 0))
        except Exception:
            logger.warning(f'Unable to convert junos secs from {input}')
            secs = 0

    delta = relativedelta(seconds=int(secs))
    return int((datetime.fromtimestamp(timestamp)-delta).timestamp()*1000)


def convert_macaddr_format_to_colon(macaddr: str) -> str:
    """Convert NXOS/EOS . macaddr form to standard : format, lowecase

    :param macaddr: str, the macaddr string to convert
    :returns: the converted macaddr string or all 0s string if arg not str
    :rtype: str

    """
    if isinstance(macaddr, str):
        if re.match(r'[0-9a-zA-Z]{4}.[0-9a-zA-Z]{4}.[0-9a-zA-Z]{4}', macaddr):
            return (':'.join([f'{x[:2]}:{x[2:]}'
                              for x in macaddr.split('.')])).lower()
        else:
            return macaddr.lower()

    return('00:00:00:00:00:00')


def convert_rangestring_to_list(rangestr: str) -> list:
    """Convert a range list such as '1, 2-5, 10, 12-20' to list
    """

    tmplst = []
    if not isinstance(rangestr, str):
        return tmplst

    try:
        for x in rangestr.split(','):
            x = x.strip().split('-')
            if x[0]:
                if len(x) == 2:
                    intrange = list(range(int(x[0]), int(x[1])+1))
                    tmplst.extend(intrange)
                else:
                    tmplst.append(int(x[0]))
    except Exception:
        logger.error(f"Range string parsing failed for {rangestr}")
        return []
    return tmplst


def convert_numlist_to_ranges(numList: List[int]) -> str:
    """Convert a given list of numbers into a range string

    Args:
        numList (List[int]): unsorted/sorted list of integers

    Returns:
        str: Range string such as '1-5, 10, 12-20'
    """
    result = ''
    for a, b in groupby(enumerate(sorted(numList)),
                        lambda pair: pair[1] - pair[0]):
        b = list(b)
        if len(b) > 1:
            result += f'{b[0][1]}-{b[-1][1]}, '
        else:
            result += f'{b[0][1]}, '

    return result[:-2]


def build_query_str(skip_fields: list, schema, ignore_regex=True,
                    **kwargs) -> str:
    """Build a pandas query string given key/val pairs
    """
    query_str = ''
    prefix = ''

    def _build_query_str(fld, val, fldtype) -> str:
        """Builds the string from the provided user input"""

        if ((fldtype == "long" or fldtype == "float") and not
                isinstance(val, str)):
            result = f'{fld} == {val}'

        elif val.startswith('!'):
            val = val[1:]
            if fldtype == "long" or fldtype == "float":
                result = f'{fld} != {val}'
            else:
                result = f'{fld} != "{val}"'
        elif val.startswith('<') or val.startswith('>'):
            result = val
        elif val.startswith('~'):
            val = val[1:]
            result = f'{fld}.str.match("{val}")'
        else:
            result = f'{fld} == "{val}"'

        return result

    for f, v in kwargs.items():
        if not v or f in skip_fields or f in ["groupby"]:
            continue
        type = schema.field(f).get('type', 'string')
        if isinstance(v, list) and len(v):
            subq = ''
            subcond = ''
            if ignore_regex and [x for x in v
                                 if isinstance(x, str) and
                                 x.startswith('~')]:
                continue

            for elem in v:
                subq += f'{subcond} {_build_query_str(f, elem, type)} '
                subcond = 'or'
            query_str += '{} ({})'.format(prefix, subq)
            prefix = "and"
        else:
            query_str += f'{prefix} {_build_query_str(f, v, type)} '
            prefix = "and"

    return query_str


def get_log_params(prog: str, cfg: dict, def_logfile: str) -> tuple:
    """Get the log file, level and size for the given program from config

    The logfile is supposed to be defined by a variable called logfile
    within the hierarchy of the config dictionary. Thus, the poller log file
    will be {'poller': {'logfile': '/tmp/sq-poller.log'}}, for example.

    :param prog: str, The name of the program. Valid values are poller, 
                      coaelscer, and rest.
    :param cfg: dict, The config dictionary
    :param def_logfile: str, The default log file to return
    :returns: log file name, log level, log size, and 
              True/False for logging to stdout
    :rtype: str, str and int

    """
    if cfg:
        logfile = cfg.get(prog, {}).get('logfile', def_logfile)
        loglevel = cfg.get(prog, {}).get('logging-level', 'WARNING')
        logsize = cfg.get(prog, {}).get('logsize', 10000000)
        log_stdout = cfg.get(prog, {}).get('log-stdout', False)
    else:
        logfile = def_logfile
        loglevel = 'WARNING'
        logsize = 10000000
        log_stdout = False

    return logfile, loglevel, logsize, log_stdout


def init_logger(logname: str,
                logfile: str,
                loglevel: str = 'WARNING',
                logsize: int = 10000000,
                use_stdout: bool = False) -> logging.Logger:
    """Initialize the logger

    :param logname: str, the name of the app that's logging
    :param logfile: str, the log file to use
    :param loglevel: str, the default log level to set the logger to
    :param use_stdout: str, log to stdout instead of or in addition to file

    """

    fh = sh = None
    # this needs to be suzieq.poller, so that it is the root of all the other pollers
    logger = logging.getLogger(logname)
    logger.setLevel(loglevel.upper())
    if logfile:
        fh = RotatingFileHandler(logfile, maxBytes=logsize, backupCount=2)
    if use_stdout:
        sh = logging.StreamHandler(sys.stdout)

    formatter = logging.Formatter(
        "%(asctime)s - %(name)s - %(levelname)s " "- %(message)s"
    )
    if fh:
        fh.setFormatter(formatter)
    if sh:
        sh.setFormatter(formatter)

    # set root logger level, so that we set asyncssh log level
    #  asynchssh sets it's level to the root level
    root = logging.getLogger()
    root.setLevel(loglevel.upper())
    if fh:
        root.addHandler(fh)
    if sh:
        root.addHandler(sh)

    logger.warning(f"log level {logging.getLevelName(logger.level)}")

    return logger


def known_devtypes() -> list:
    """Returns the list of known dev types"""
    return(['cumulus', 'eos', 'iosxe', 'iosxr', 'ios', 'junos-mx', 'junos-qfx',
            'junos-ex', 'junos-es', 'linux', 'nxos', 'sonic'])


def humanize_timestamp(field: pd.Series, tz=None) -> pd.Series:
    '''Convert the UTC timestamp in Dataframe to local time.
    Use of pd.to_datetime will not work as it converts the timestamp
    to UTC. If the timestamp is already in UTC format, we get busted time.
    '''
    if field.empty:
        return field

    tz = tz or get_localzone().zone
    return field.apply(lambda x: datetime.utcfromtimestamp((int(x)/1000))) \
                .dt.tz_localize('UTC').dt.tz_convert(tz)


def expand_nxos_ifname(ifname: str) -> str:
    '''Expand shortned ifnames in NXOS to their full values, if required'''
    if not ifname:
        return ''
    if ifname.startswith('Eth') and 'Ether' not in ifname:
        return ifname.replace('Eth', 'Ethernet')
    elif ifname.startswith('Po') and 'port' not in ifname:
        return ifname.replace('Po', 'port-channel')
    return ifname


def expand_eos_ifname(ifname: str) -> str:
    '''Expand shortned ifnames in EOS to their full values, if required'''
    if not ifname:
        return ''
    if ifname.startswith('Eth') and 'Ether' not in ifname:
        return ifname.replace('Eth', 'Ethernet')
    elif ifname.startswith('Po') and 'Port' not in ifname:
        return ifname.replace('Po', 'Port-Channel')
    elif ifname.startswith('Vx') and 'Vxlan' not in ifname:
        return ifname.replace('Vx', 'Vxlan')
    return ifname


def ensure_single_instance(filename: str, block: bool = False) -> int:
    """Check there's only a single active instance of a process using lockfile

    It optionally can block waiting for the resource the become available.

    Use a pid file with advisory file locking to assure this.

    :returns: fd if lock was successful or 0
    :rtype: int

    """
    basedir = os.path.dirname(filename)
    if not os.path.exists(basedir):
        # Permission error or any other error will abort
        os.makedirs(basedir, exist_ok=True)

    fd = os.open(filename, os.O_RDWR | os.O_CREAT, 0o600)
    if fd:
        try:
            if block:
                fcntl.flock(fd, fcntl.LOCK_EX)
            else:
                fcntl.flock(fd, fcntl.LOCK_EX | fcntl.LOCK_NB)
            os.truncate(fd, 0)
            os.write(fd, bytes(str(os.getpid()), 'utf-8'))
        except OSError:
            if OSError.errno == errno.EBUSY:
                # Return the PID of the process thats locked the file
                bpid = os.read(fd, 10)
                os.close(fd)
                try:
                    fd = -int(bpid)
                except ValueError:
                    fd = 0
            else:
                os.close(fd)
                fd = 0

    return fd


def expand_ios_ifname(ifname: str) -> str:
    """Get expanded interface name for IOSXR/XE given its short form

    :param ifname: str, short form of IOSXR interface name
    :returns: Expanded version of short form interface name
    :rtype: str
    """

    ifmap = {'BE': 'Bundle-Ether',
             'BV': 'BVI',
             'Fi': 'FiftyGigE',
             'Fo': 'FortyGigE',
             'FH': 'FourHundredGigE',
             'Gi': 'GigabitEthernet',
             'Gig': 'GigabitEthernet',
             'Hu': 'HundredGigE',
             'Lo': 'Loopback',
             'Mg': 'MgmtEth',
             'Nu': 'Null',
             'TE': 'TenGigE',
             'TF': 'TwentyFiveGigE',
             'TH': 'TwoHundredGigE',
             'tsec': 'tunnel-ipsec',
             'tmte': 'tunnel-mte',
             'tt': 'tunnel-te',
             'tp': 'tunnel-tp',
             'Vl': 'Vlan',
             'CPU': 'cpu',
             }
    pfx = re.match(r'[a-zA-Z]+', ifname)
    if pfx:
        pfxstr = pfx.group(0)
        if pfxstr in ifmap:
            return ifname.replace(pfxstr, ifmap[pfxstr])

    return ifname


def get_sq_install_dir() -> str:
    '''Return the absolute path of the suzieq installation dir'''
    spec = find_spec('suzieq')
    if spec:
        return(os.path.dirname(spec.loader.path))
    else:
        return(os.path.abspath('./'))


def get_sleep_time(period: str) -> int:
    """Returns the duration in seconds to sleep given a period

    Checking if the period format matches a specified format MUST be
    done by the caller.

    :param period: str, the period of form <value><unit>, '15m', '1h' etc
    :returns: duration to sleep in seconds
    :rtype: int
    """
    tm, unit, _ = re.split(r'(\D)', period)
    now = datetime.now()
    nextrun = parse(period, settings={'PREFER_DATES_FROM': 'future'})
    if unit == 'm':
        nextrun = nextrun.replace(second=0)
    elif unit == 'h':
        nextrun = nextrun.replace(minute=0, second=0)
    else:
        nextrun = nextrun.replace(hour=0, minute=0, second=0)

    return (nextrun-now).seconds



from ipaddress import IPv4Network
import pandas as pd
import numpy as np

from .engineobj import SqPandasEngine
from suzieq.utils import SchemaForTable, build_query_str, humanize_timestamp


class OspfObj(SqPandasEngine):

    @staticmethod
    def table_name():
        return 'ospf'

    def _get_combined_df(self, **kwargs):
        """OSPF has info divided across multiple tables. Get a single one"""

        columns = kwargs.pop('columns', ['default'])
        state = kwargs.pop('state', '')
        addnl_fields = kwargs.pop('addnl_fields', self.iobj._addnl_fields)
        addnl_nbr_fields = self.iobj._addnl_nbr_fields
        user_query = kwargs.pop('query_str', '')
        hostname = kwargs.pop('hostname', [])

        cols = self.schema.get_display_fields(columns)
        if columns == ['default']:
            cols.append('timestamp')

        ifschema = SchemaForTable('ospfIf', schema=self.all_schemas)
        nbrschema = SchemaForTable('ospfNbr', schema=self.all_schemas)

        if (columns != ['default']) and (columns != ['*']):
            ifkeys = ifschema.key_fields()
            nbrkeys = nbrschema.key_fields()
            if_flds = ifschema.fields
            nbr_flds = nbrschema.fields

            ifcols = ifkeys
            nbrcols = nbrkeys
            for fld in columns:
                if fld in if_flds and fld not in ifcols:
                    ifcols.append(fld)
                elif fld in nbr_flds and fld not in nbrcols:
                    nbrcols.append(fld)
        else:
            ifcols = ifschema.get_display_fields(columns)
            nbrcols = nbrschema.get_display_fields(columns)

        state_query_dict = {
            'full': '(adjState == "full" or adjState == "passive")',
            'passive': '(adjState == "passive")',
            'other': '(adjState != "full" and adjState != "passive")',
            '!full': '(adjState != "full")',
            '!passive': '(adjState != "passive")',
            '!other': '(adjState == "full" or adjState == "passive")',
        }

        if state:
            query_str = state_query_dict.get(state, '')
            cond_prefix = ' and '
        else:
            query_str = ''
            cond_prefix = ''

        host_query_str = build_query_str([], ifschema, hostname=hostname)
        if host_query_str:
            query_str += f'{cond_prefix}{host_query_str}'

        df = self.get_valid_df('ospfIf', addnl_fields=addnl_fields,
                               columns=ifcols, **kwargs)
        nbr_df = self.get_valid_df('ospfNbr', addnl_fields=addnl_nbr_fields,
                                   columns=nbrcols, **kwargs)
        if nbr_df.empty:
            return df

        merge_cols = [x for x in ['namespace', 'hostname', 'ifname']
                      if x in nbr_df.columns]
        # Merge the two tables
        df = df.merge(nbr_df, on=merge_cols, how='left')

        # This is because some NOS have the ipAddress in nbr table and some in
        # interface table. Nbr table wins over interface table if present
        if 'ipAddress_y' in df:
            df['ipAddress'] = np.where(
                df['ipAddress_y'] == "",
                df['ipAddress_x'], df['ipAddress_y'])
            df['ipAddress'] = np.where(df['ipAddress'], df['ipAddress'],
                                       df['ipAddress_x'])

        if columns == ['*']:
            df = df.drop(columns=['area_y', 'instance_y', 'vrf_y',
                                  'ipAddress_x', 'ipAddress_y', 'areaStub_y',
                                  'timestamp_y'], errors='ignore') \
                .rename(columns={
                    'instance_x': 'instance', 'areaStub_x': 'areaStub',
                    'area_x': 'area', 'vrf_x': 'vrf',
                    'state_x': 'ifState', 'state_y': 'adjState',
                    'sqvers_x': 'sqvers', 'active_x': 'active',
                    'timestamp_x': 'timestamp'})
        else:
            df = df.rename(columns={'vrf_x': 'vrf', 'area_x': 'area',
                                    'state_x': 'ifState',
                                    'state_y': 'adjState',
                                    'timestamp_x': 'timestamp'})
            df = df.drop(list(df.filter(regex='_y$')), axis=1) \
                   .drop(['ipAddress_x'], axis=1, errors='ignore') \
                   .fillna({'peerIP': '-', 'numChanges': 0,
                            'lastChangeTime': 0})

        if df.empty:
            return df

        if 'lastChangeTime' in df.columns:
            df['lastChangeTime'] = np.where(df.lastChangeTime == '-',
                                            0, df.lastChangeTime)
        # Fill the adjState column with passive if passive
        if 'passive' in df.columns:
            df.loc[df['adjState'].isnull(), 'adjState'] = df['passive']
            df.loc[df['adjState'].eq(True), 'adjState'] = 'passive'
            df.loc[df['adjState'].eq(False), 'adjState'] = 'fail'
            df.loc[df['adjState'] == 'passive', 'peerIP'] = ''
            df.loc[df['adjState'] == 'passive', 'peerRouterId'] = ''

            df.drop(columns=['passive'], inplace=True)

        df.bfill(axis=0, inplace=True)

        if 'peerHostname' in columns or (columns in [['*'], ['default']]):
            nfdf = df.query('adjState != "full"').reset_index()
            nfdf['peerHostname'] = ''
            newdf = df.query('adjState == "full"').reset_index() \
                .drop('peerHostname', axis=1, errors='ignore')
            if not newdf.empty:
                newdf['matchIP'] = newdf.ipAddress.str.split('/').str[0]
                newdf = newdf.merge(newdf[['namespace', 'hostname', 'vrf',
                                           'matchIP']],
                                    left_on=['namespace', 'vrf', 'peerIP'],
                                    right_on=['namespace', 'vrf', 'matchIP'],
                                    suffixes=["", "_y"]) \
                    .rename(columns={'hostname_y': 'peerHostname'}) \
                    .drop_duplicates(subset=['namespace', 'hostname',
                                             'vrf', 'ifname']) \
                    .drop(columns=['matchIP', 'matchIP_y', 'timestamp_y'],
                          errors='ignore')

                if newdf.empty:
                    newdf = df.query('adjState == "full"').reset_index()
                    newdf['peerHostname'] = ''
                final_df = pd.concat([nfdf, newdf])
            else:
                final_df = df
        else:
            final_df = df.drop(list(df.filter(regex='_y$')), axis=1) \
                         .rename({'timestamp_x': 'timestamp'})

        if query_str:
            final_df = final_df.query(query_str).reset_index(drop=True)

        if user_query and not final_df.empty:
            final_df = self._handle_user_query_str(final_df, user_query)
        # Move the timestamp column to the end
        return final_df[cols]

    def get(self, **kwargs):
        return self._get_combined_df(**kwargs)

    def summarize(self, **kwargs):
        """Describe the data"""

        # Discard these
        kwargs.pop('columns', None)

        # 'ospfIf' is ignored
        self._init_summarize('ospfIf', **kwargs)
        if self.summary_df.empty:
            return self.summary_df

        self._summarize_on_add_field = [
            ('deviceCnt', 'hostname', 'nunique'),
            ('peerCnt', 'hostname', 'count'),
        ]

        self._summarize_on_add_with_query = [
            ('stubbyPeerCnt', 'areaStub', 'areaStub'),
            ('passivePeerCnt', 'adjState == "passive"', 'ifname'),
            ('unnumberedPeerCnt', 'isUnnumbered', 'isUnnumbered'),
            ('failedPeerCnt', 'adjState != "passive" and nbrCount == 0',
             'ifname'),
        ]

        self._summarize_on_add_list_or_count = [
            ('area', 'area'),
            ('vrf', 'vrf'),
            ('helloTime', 'helloTime'),
            ('deadTime', 'deadTime'),
            ('retxTime', 'retxTime'),
            ('networkType', 'networkType'),
        ]

        self.summary_df['lastChangeTime'] = np.where(
            self.summary_df.lastChangeTime.isnull(), 0,
            self.summary_df.lastChangeTime)

        self.summary_df['lastChangeTime'] = humanize_timestamp(
            self.summary_df.lastChangeTime, self.cfg.get('analyzer', {})
            .get('timezone', None))

        self.summary_df['lastChangeTime'] = (
            self.summary_df['timestamp'] - self.summary_df['lastChangeTime'])
        self.summary_df['lastChangeTime'] = self.summary_df['lastChangeTime'] \
            .apply(lambda x: x.round('s'))

        self._summarize_on_add_stat = [
            ('adjChangesStat', '', 'numChanges'),
            ('upTimeStat', 'adjState == "full"', 'lastChangeTime'),
        ]

        self._gen_summarize_data()
        self._post_summarize()
        return self.ns_df.convert_dtypes()

    def aver(self, **kwargs):
        """Assert that the OSPF state is OK"""

        kwargs.pop('columns', [])
        columns = [
            "namespace",
            "hostname",
            "vrf",
            "ifname",
            "state",
            "routerId",
            "helloTime",
            "deadTime",
            "passive",
            "ipAddress",
            "isUnnumbered",
            "areaStub",
            "networkType",
            "timestamp",
            "area",
            "nbrCount",
        ]

        status = kwargs.pop('status', 'all')
        # we have to not filter hostname at this point because we need to
        #   understand neighbor relationships

        ospf_df = self.get_valid_df("ospfIf", columns=columns,
                                    state="!adminDown", **kwargs)
        if ospf_df.empty:
            return pd.DataFrame(columns=columns)

        ospf_df["assertReason"] = [[] for _ in range(len(ospf_df))]
        df = (
            ospf_df[ospf_df["routerId"] != ""]
            .groupby(["routerId", "namespace"], as_index=False)[["hostname",
                                                                 "namespace"]]
            .agg(lambda x: x.unique().tolist())
        ).dropna(how='any')

        # df is a dataframe with each row containing the routerId and the
        # corresponding list of hostnames with that routerId. In a good
        # configuration, the list must have exactly one entry.
        if not df.empty:
            # This check is because we don't get routerID with Arista boxes
            ospf_df['assertReason'] = (
                ospf_df.merge(df, on=["routerId"], how="outer")
                .apply(lambda x: ["duplicate routerId {}".format(
                    x["hostname_y"])]
                    if x["hostname_y"] is not np.nan and len(x['hostname_y']) != 1 else [], axis=1))

        # Now  peering match
        lldp_df = self._get_table_sqobj('lldp').get(
            namespace=kwargs.get('namespace', []), columns=['default'])
        if lldp_df.empty:
            ospf_df = ospf_df[~(ospf_df.ifname.str.contains('loopback') |
                                ospf_df.ifname.str.contains('Vlan'))]
            ospf_df['assertReason'] = 'No LLDP peering info'
            ospf_df['assert'] = 'fail'
            return ospf_df[['namespace', 'hostname', 'vrf', 'ifname',
                            'assertReason', 'assert']]

        # Create a single massive DF with fields populated appropriately
        use_cols = [
            "namespace",
            "routerId",
            "hostname",
            "vrf",
            "ifname",
            "helloTime",
            "deadTime",
            "passive",
            "ipAddress",
            "areaStub",
            "isUnnumbered",
            "networkType",
            "area",
            "timestamp",
            "lldpIfname"
        ]

        # In case of Junos, the OSPF interface name is a subif of
        # the interface name. So, create a new column with the
        # orignal interface name.
        ospf_df['lldpIfname'] = ospf_df['ifname'].str.split('.').str[0]
        int_df = ospf_df[use_cols].merge(
            lldp_df,
            left_on=["namespace", "hostname", "lldpIfname"],
            right_on=['namespace', 'hostname', "ifname"],
            suffixes=("", "_y")) \
            .dropna(how="any")

        if int_df.empty:
            # Weed out the loopback and SVI interfaces as they have no LLDP peers
            if status == "pass":
                ospf_df = ospf_df[(ospf_df.ifname.str.contains('loopback') |
                                   ospf_df.ifname.str.contains('Vlan'))]
                ospf_df['assertReason'] = []
                ospf_df['assert'] = 'pass'
            else:
                ospf_df = ospf_df[~(ospf_df.ifname.str.contains('loopback') |
                                    ospf_df.ifname.str.contains('Vlan'))]
                ospf_df['assertReason'] = 'No LLDP peering info'
                ospf_df['assert'] = 'fail'

            return ospf_df[['namespace', 'hostname', 'vrf', 'ifname',
                            'assertReason', 'assert']]

        peer_df = ospf_df.merge(
            int_df,
            left_on=["namespace", "hostname", "lldpIfname"],
            right_on=["namespace", "peerHostname", "peerIfname"]) \
            .dropna(how="any")

        if peer_df.empty:
            ospf_df = ospf_df[~(ospf_df.ifname.str.contains('loopback') |
                                ospf_df.ifname.str.contains('Vlan'))]
            ospf_df['assertReason'] = 'No LLDP peering info'
            ospf_df['assert'] = 'fail'
        else:
            ospf_df = peer_df
            # Now start comparing the various parameters
            ospf_df["assertReason"] += ospf_df.apply(
                lambda x: ["subnet mismatch"]
                if (
                    (x["isUnnumbered_x"] != x["isUnnumbered_y"])
                    and (
                        IPv4Network(x["ipAddress_x"], strict=False)
                        != IPv4Network(x["ipAddress_y"], strict=False)
                    )
                )
                else [],
                axis=1,
            )
            ospf_df["assertReason"] += ospf_df.apply(
                lambda x: ["area mismatch"]
                if (x["area_x"] != x["area_y"]) else [], axis=1)

            ospf_df["assertReason"] += ospf_df.apply(
                lambda x: ["area stub mismatch"]
                if (x["areaStub_x"] != x["areaStub_y"]) else [], axis=1)

            ospf_df["assertReason"] += ospf_df.apply(
                lambda x: ["Hello timers mismatch"]
                if x["helloTime_x"] != x["helloTime_y"]
                else [],
                axis=1,
            )
            ospf_df["assertReason"] += ospf_df.apply(
                lambda x: ["Dead timer mismatch"]
                if x["deadTime_x"] != x["deadTime_y"]
                else [],
                axis=1,
            )
            ospf_df["assertReason"] += ospf_df.apply(
                lambda x: ["network type mismatch"]
                if x["networkType_x"] != x["networkType_y"]
                else [],
                axis=1,
            )
            ospf_df["assertReason"] += ospf_df.apply(
                lambda x: ["passive config mismatch"]
                if x["passive_x"] != x["passive_y"]
                else [],
                axis=1,
            )
            ospf_df["assertReason"] += ospf_df.apply(
                lambda x: ["vrf mismatch"] if x["vrf_x"] != x["vrf_y"] else [],
                axis=1,
            )

        # Fill up a single assert column now indicating pass/fail
        ospf_df['assert'] = ospf_df.apply(lambda x: 'pass'
                                          if not len(x['assertReason'])
                                          else 'fail', axis=1)

        result = (
            ospf_df.rename(
                index=str,
                columns={
                    "hostname_x": "hostname",
                    "ifname_x": "ifname",
                    "vrf_x": "vrf",
                    "timestamp_x": "timestamp",
                },
            )[["namespace", "hostname", "ifname", "vrf", "assert",
               "assertReason", "timestamp"]].explode(column='assertReason')
            .fillna({'assertReason': '-'})
        )

        if status == "pass":
            return result.query('assertReason == "-"')
        elif status == "fail":
            return result.query('assertReason != "-"')

        return result.reset_index(drop=True)



from suzieq.sqobjects.basicobj import SqObject
import pandas as pd
from suzieq.utils import humanize_timestamp


class InventoryObj(SqObject):
    def __init__(self, **kwargs):
        super().__init__(table='inventory', **kwargs)
        self._valid_get_args = ['namespace', 'hostname', 'columns', 'type',
                                'serial', 'vendor', 'status', 'model',
                                'query_str']



import numpy as np
from suzieq.utils import expand_ios_ifname
from suzieq.poller.services.service import Service


class VlanService(Service):
    """Vlan service. Different class because Vlan is not right type for EOS"""

    def _clean_eos_data(self, processed_data, raw_data):
        '''Massage the interface output'''

        for entry in processed_data:
            if (entry['vlanName'].startswith('VLAN') or
                    entry['vlanName'] == "default"):
                entry['vlanName'] = f'vlan{entry["vlan"]}'

        return processed_data

    def _clean_cumulus_data(self, processed_data, raw_data):
        '''Fix Linux VLAN output.
        Linux output is the transposed from the vlan output of all other NOS
        '''

        new_entries = []
        entry_dict = {}

        for entry in processed_data:

            for item in entry['vlan']:
                vlan = int(item)
                vlanName = f'vlan{vlan}'

                if entry['vlanName'] == 'bridge':
                    state = 'suspended'
                else:
                    state = 'active'
                if vlanName not in entry_dict:
                    new_entry = {'vlanName': vlanName,
                                 'state': state,
                                 'vlan': vlan,
                                 'interfaces': set(),
                                 }
                    new_entries.append(new_entry)
                    entry_dict[vlanName] = new_entry
                else:
                    new_entry = entry_dict[vlanName]

                if entry['vlanName'] != 'bridge':
                    new_entry['interfaces'].add(entry['vlanName'])
                    new_entry['state'] = 'active'

        for entry in new_entries:
            entry['interfaces'] = list(entry['interfaces'])

        return new_entries

    def _clean_sonic_data(self, processed_data, raw_data):
        return self._clean_cumulus_data(processed_data, raw_data)

    def _clean_nxos_data(self, processed_data, raw_data):
        '''Massage the interface output'''

        for entry in processed_data:
            if (entry['vlanName'].startswith('VLAN') or
                    entry['vlanName'] == "default"):
                entry['vlanName'] = f'vlan{entry["vlan"]}'
            if isinstance(entry['interfaces'], str):
                entry['interfaces'] = entry['interfaces'].split(',')
            else:
                entry['interfaces'] = entry['interfaces'][0].split(',')
        return processed_data

    def _clean_junos_data(self, processed_data, raw_data):
        '''Massage the default name and interface list'''

        drop_indices = []

        for i, entry in enumerate(processed_data):
            if entry['vlan'] is None:
                drop_indices.append(i)
                continue

            if entry['vlanName'] == 'default':
                entry['vlanName'] = f'vlan{entry["vlan"]}'
            if entry['interfaces'] == [[None]]:
                entry['interfaces'] = []
            # We don't need the explicit .<vlan> tag for interfaces
            # to keep it consistent with the other devices, but we
            # cannot remove the VTEP info
            entry['interfaces'] = [x.split('.')[0].replace('*', '')
                                   if not x.startswith('vtep')
                                   else x.replace('*', '')
                                   for x in entry['interfaces']]
            entry['state'] = entry['state'].lower()
            entry['vlanName'] = entry['vlanName'].lower()

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_ios_data(self, processed_data, raw_data):
        '''Massage the interface list'''

        for entry in processed_data:
            if entry['vlanName'] == 'default':
                entry['vlanName'] = f'vlan{entry["vlan"]}'
            if entry['interfaces']:
                newiflist = []
                for ifname in entry['interfaces'].split(','):
                    newiflist.append(expand_ios_ifname(ifname.strip()))
                entry['interfaces'] = newiflist
            entry['state'] = entry['state'].lower()

        return processed_data

    def _clean_iosxe_data(self, processed_data, raw_data):
        return self._clean_ios_data(processed_data, raw_data)



import time
import pandas as pd
from nubia import command, argument

from suzieq.cli.sqcmds.command import SqCommand, ArgHelpClass
from suzieq.sqobjects.path import PathObj


@command("path", help="build and act on path data")
class PathCmd(SqCommand):
    """Path trace information including overlay and underlay"""

    @argument("src", description="Source IP address, in quotes")
    @argument("dest", description="Destination IP address, in quotes")
    @argument("vrf", description="VRF to trace path in")
    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        columns: str = "default",
        src: str = "",
        dest: str = "",
        vrf: str = "",
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            sqobj=PathObj
        )
        self.src = src
        self.dest = dest
        self.vrf = vrf
        self._additional_help_vars = {
            # These are only required for args defined only in this class def
            'src': ArgHelpClass('Source IP address, in quotes'),
            'dest': ArgHelpClass('Destination IP address, in quotes'),
            'vrf': ArgHelpClass('VRF to trace path in')
        }

    @command("show")
    def show(self):
        """Show paths between specified from source to target ip addresses"""
        # Get the default display field names
        if self.columns is None:
            return

        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        try:
            df = self.sqobj.get(
                hostname=self.hostname, columns=self.columns,
                namespace=self.namespace, source=self.src, dest=self.dest,
                vrf=self.vrf
            )
        except Exception as e:
            if len(e.args) > 1:
                df = e.args[1]  # The second arg if present is the path so far
                df['error'] = str(e.args[0])
            else:
                df = pd.DataFrame(
                    {'error': ['ERROR: {}'.format(str(e.args[0]))]})

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if not df.empty:
            return self._gen_output(df, sort=False)

    @command("summarize")
    def summarize(self):
        """Summarize paths between specified from source to target ip addresses"""
        # Get the default display field names
        if self.columns is None:
            return

        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        try:
            df = self.sqobj.summarize(
                hostname=self.hostname, namespace=self.namespace,
                source=self.src, dest=self.dest, vrf=self.vrf
            )
        except Exception as e:
            df = pd.DataFrame({'error': ['ERROR: {}'.format(str(e))]})

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if not df.empty:
            return self._gen_output(df)

    @command("unique")
    @argument("count", description="include count of times a value is seen",
              choices=['True'])
    def unique(self, count: str = 'False'):
        """Display unique values (and counts) for specified field of a path"""
        # Get the default display field names
        if self.columns is None:
            return

        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        try:
            df = self.sqobj.unique(
                hostname=self.hostname, namespace=self.namespace,
                source=self.src, dest=self.dest, vrf=self.vrf,
                count=count
            )
        except Exception as e:
            df = pd.DataFrame({'error': ['ERROR: {}'.format(str(e))]})

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if not df.empty:
            return self._gen_output(df)

    @command("top", help="find the top n values for a field")
    @argument("count", description="number of rows to return")
    @argument("what", description="integer field to get top values for")
    @argument("reverse", description="return bottom n values",
              choices=['True', 'False'])
    def top(self, count: int = 5, what: str = '', reverse: str = 'False',
            **kwargs) -> int:
        """Return the top n values for a field in path trace output

        Args:
            count (int, optional): The number of entries to return. Defaults to 5
            what (str, optional): Field name to use for largest/smallest val
            reverse (bool, optional): Reverse and return n smallest

        Returns:
            int: 0 or error code
        """
        now = time.time()

        df = self.sqobj.top(hostname=self.hostname,
                            namespace=self.namespace,
                            what=what, count=count, reverse=eval(reverse),
                            source=self.src, dest=self.dest, vrf=self.vrf,
                            )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if 'error' in df.columns:
            return self._gen_output(df)

        if not df.empty:
            df = self.sqobj.humanize_fields(df)
            return self._gen_output(df.sort_values(by=[what], ascending=False),
                                    dont_strip_cols=True, sort=False)
        else:
            return self._gen_output(df)



from datetime import datetime, timedelta, timezone
from ipaddress import ip_address
from dateparser import parse

import numpy as np

from suzieq.poller.services.service import Service
from suzieq.utils import get_timestamp_from_cisco_time
from suzieq.utils import get_timestamp_from_junos_time


class OspfNbrService(Service):
    """OSPF Neighbor service. Output needs to be munged"""

    def frr_convert_reltime_to_epoch(self, reltime, timestamp):
        """Convert string of type 1d12h3m23s into absolute epoch"""
        secs = 0
        s = reltime
        if not reltime:
            return 0

        for t, mul in {
            "w": 3600 * 24 * 7,
            "d": 3600 * 24,
            "h": 3600,
            "m": 60,
            "s": 1,
        }.items():
            v = s.split(t)
            if len(v) == 2:
                secs += int(v[0]) * mul
            s = v[-1]

        return int((timestamp/1000) - secs) * 1000

    def _clean_linux_data(self, processed_data, raw_data):

        if not raw_data:
            return processed_data

        if isinstance(raw_data, list):
            read_from = raw_data[0]
        else:
            read_from = raw_data
        timestamp = read_from["timestamp"]

        for entry in processed_data:
            entry["vrf"] = "default"
            entry["state"] = entry["state"].lower()
            entry["lastUpTime"] = self.frr_convert_reltime_to_epoch(
                entry["lastUpTime"], timestamp
            )
            entry["lastDownTime"] = self.frr_convert_reltime_to_epoch(
                entry["lastDownTime"], timestamp
            )
            if entry["lastUpTime"] > entry["lastDownTime"]:
                entry["lastChangeTime"] = entry["lastUpTime"]
            else:
                entry["lastChangeTime"] = entry["lastDownTime"]
            entry["areaStub"] = entry["areaStub"] == "[Stub]"
            if not entry["bfdStatus"]:
                entry["bfdStatus"] = "disabled"
            else:
                entry["bfdStatus"] = entry['bfdStatus'].lower()

        return processed_data

    def _clean_cumulus_data(self, processed_data, raw_data):
        return self._clean_linux_data(processed_data, raw_data)

    def _clean_sonic_data(self, processed_data, raw_data):
        return self._clean_linux_data(processed_data, raw_data)

    def _clean_eos_data(self, processed_data, raw_data):
        for entry in processed_data:
            entry["state"] = entry["state"].lower()
            entry["lastChangeTime"] = int(entry["lastChangeTime"] * 1000)
            # What is provided is the opposite of stub and so we not it
            entry["areaStub"] = not entry["areaStub"]

            bfd_status = entry.get("bfdStatus", '')
            if not bfd_status or (bfd_status == 'adminDown'):
                entry["bfdStatus"] = "disabled"
            else:
                entry["bfdStatus"] = bfd_status.lower()

        return processed_data

    def _clean_junos_data(self, processed_data, raw_data):

        ifentries = {}
        drop_indices = []

        for i, entry in enumerate(processed_data):
            if entry.get('_entryType', '') == '_bfdType':
                ifname = entry.get('ifname', '')
                if ifentries.get(ifname, {}):
                    if any('OSPF' in x for x in entry['_client']):
                        ifentry = ifentries[ifname]
                        ifentry['bfdStatus'] = entry['bfdStatus'].lower()
                drop_indices.append(i)
                continue

            vrf = entry['vrf'][0]['data']
            if vrf == "master":
                entry['vrf'] = "default"
            else:
                entry['vrf'] = vrf

            entry['lastChangeTime'] = get_timestamp_from_junos_time(
                entry['lastChangeTime'], raw_data[0]['timestamp']/1000)
            entry['state'] = entry['state'].lower()
            entry['bfdStatus'] = 'disabled'
            ifentries[entry['ifname']] = entry

        processed_data = np.delete(processed_data, drop_indices).tolist()
        return processed_data

    def _clean_nxos_data(self, processed_data, raw_data):
        for entry in processed_data:
            entry['state'] = entry['state'].lower()
            entry['numChanges'] = int(entry['numChanges'])
            # Cisco's format examples are PT7H28M21S, P1DT4H9M46S
            entry['lastChangeTime'] = get_timestamp_from_cisco_time(
                entry['lastChangeTime'], raw_data[0]['timestamp']/1000)

            if not entry.get("bfdStatus", ''):
                entry["bfdStatus"] = "disabled"
            else:
                entry["bfdStatus"] = entry['bfdStatus'].lower()
        return processed_data

    def _clean_ios_data(self, processed_data, raw_data):
        for entry in processed_data:
            # make area the dotted model
            area = entry.get('area', '')
            if area.isdecimal():
                entry['area'] = str(ip_address(int(area)))
            entry['state'] = entry['state'].lower()
            entry['lastUpTime'] = parse(entry['lastUpTime']).timestamp()
            entry['lastChangeTime'] = int(entry['lastUpTime'])*1000
            entry['lastDownTime'] = 0
            entry['lsaRtxCnt'] = int(entry['lsaRetxCnt'])
            entry['areaStub'] = entry['areaStub'] == 'Stub'
            entry['vrf'] = "default"
            entry['nbrPrio'] = int(entry['nbrPrio']) if entry['nbrPrio'] else 0
            if not entry.get("bfdStatus", ''):
                entry["bfdStatus"] = "unknown"
            else:
                entry["bfdStatus"] = entry['bfdStatus'].lower()

        return processed_data

    def _clean_iosxe_data(self, processed_data, raw_data):
        return self._clean_ios_data(processed_data, raw_data)



import time
from nubia import command, argument
import pandas as pd

from suzieq.cli.sqcmds.command import SqCommand
from suzieq.sqobjects.interfaces import IfObj
from suzieq.utils import humanize_timestamp


@command("interface", help="Act on Interface data", aliases=['interfaces'])
class InterfaceCmd(SqCommand):
    """Device interface information including MTU, Speed, IP address etc"""

    def __init__(
        self,
        engine: str = "",
        hostname: str = "",
        start_time: str = "",
        end_time: str = "",
        view: str = "",
        namespace: str = "",
        format: str = "",
        columns: str = "default",
        query_str: str = ' ',
    ) -> None:
        super().__init__(
            engine=engine,
            hostname=hostname,
            start_time=start_time,
            end_time=end_time,
            view=view,
            namespace=namespace,
            columns=columns,
            format=format,
            query_str=query_str,
            sqobj=IfObj
        )

    @command("show")
    @argument("ifname", description="interface name to qualify")
    @argument("type", description="interface type to qualify")
    @argument("state", description="interface state to qualify show",
              choices=["up", "down", "notConnected"])
    @argument("mtu", description="filter interfaces with MTU")
    @argument("vrf", description="filter interfaces matching VRFs")
    def show(self, ifname: str = "", state: str = "", type: str = "",
             mtu: str = "", vrf: str = "") -> None:
        """Show interface info
        """
        if self.columns is None:
            return

        # Get the default display field names
        now = time.time()
        if self.columns != ["default"]:
            self.ctxt.sort_fields = None
        else:
            self.ctxt.sort_fields = []

        df = self._invoke_sqobj(self.sqobj.get,
                                hostname=self.hostname,
                                ifname=ifname.split(),
                                columns=self.columns,
                                namespace=self.namespace,
                                state=state,
                                mtu=mtu.split(),
                                vrf=vrf.split(),
                                query_str=self.query_str,
                                type=type.split(),
                                )
        if 'statusChangeTimestamp' in df.columns:
            df['statusChangeTimestamp'] = humanize_timestamp(
                df.statusChangeTimestamp,
                self.cfg.get('analyzer', {}).get('timezone', None))

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df)

    @command("unique", help="find the list of unique items in a column")
    @argument("count", description="include count of times a value is seen",
              choices=['True'])
    @argument("type", description="include type of ports to include")
    def unique(self, count: str = '', type: str = '', **kwargs):
        """Get unique values (and counts) associated with requested field"""
        now = time.time()

        df = self._invoke_sqobj(self.sqobj.unique,
                                hostname=self.hostname,
                                namespace=self.namespace,
                                query_str=self.query_str,
                                type=type.split(),
                                addnl_fields=['type'],
                                count=count,
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        if 'error' in df.columns:
            return self._gen_output(df)

        if not count or df.empty:
            return self._gen_output(df.sort_values(by=[self.columns[0]]),
                                    dont_strip_cols=True)
        else:
            return self._gen_output(
                df.sort_values(by=['numRows', self.columns[0]]),
                dont_strip_cols=True)

    @command("assert")
    @argument("ifname", description="interface name to qualify")
    @argument(
        "what",
        description="What do you want to assert",
        choices=["mtu-value"],
    )
    @argument("value", description="Value to match against")
    @argument("status", description="Show only assert that matches this value",
              choices=["all", "fail", "pass"])
    @argument("ignore_missing_peer",
              description="Treat missing peer as passing assert check",
              choices=["True", "False"])
    def aver(self, ifname: str = "", state: str = "", what: str = "",
             value: str = '', status: str = 'all',
             ignore_missing_peer: str = "False"):
        """Assert aspects about the interface
        """
        if self.columns != ["default"]:
            df = pd.DataFrame(
                {'error': ['ERROR: You cannot specify columns with assert']})
            return self._gen_output(df)

        now = time.time()

        if what == "mtu-value" and value == '':
            print("Provide value to match MTU against")
            return
        elif not what:
            value = ''

        df = self._invoke_sqobj(self.sqobj.aver,
                                hostname=self.hostname,
                                ifname=ifname.split(),
                                namespace=self.namespace,
                                what=what,
                                matchval=value.split(),
                                status=status,
                                ignore_missing_peer=(
                                    ignore_missing_peer == "True"),
                                )

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)

        return self._assert_gen_output(df)

    @command("top", help="Top for any given field")
    @argument("what", description="Field you want to see top for")
    @argument("count", description="How many top entries")
    @argument("reverse", description="True see Bottom n",
              choices=["True", "False"])
    @argument("type", description="interface type to qualify")
    @argument('ifname', description="Interface name to qualify")
    def top(self, what: str = "", count: int = 5, reverse: str = "False",
            ifname: str = '', type: str = 'ethernet'):
        """Show top n entries based on specific field
        """
        if self.columns is None:
            return

        now = time.time()

        df = self._invoke_sqobj(self.sqobj.top,
                                hostname=self.hostname,
                                what=what,
                                count=count,
                                ifname=ifname.split(),
                                reverse=(reverse == "True") or False,
                                type=type.split(),        # phy interfaces only
                                columns=self.columns,
                                query_str=self.query_str,
                                namespace=self.namespace,
                                )

        if 'statusChangeTimestamp' in df.columns:
            df['statusChangeTimestamp'] = humanize_timestamp(
                df.statusChangeTimestamp,
                self.cfg.get('analyzer', {}).get('timezone', None))

        self.ctxt.exec_time = "{:5.4f}s".format(time.time() - now)
        return self._gen_output(df, sort=False)



from dataclasses import dataclass
from importlib.util import find_spec
import os
from pathlib import Path

import streamlit as st


@dataclass
class HelpSessionState:
    help_on: str = ''


def get_title():
    # suzieq_gui.py has hardcoded this name.
    return '_Help_'


def help_sidebar(state: HelpSessionState, helpdir: str):
    '''Sidebar page for help'''

    filelist = Path(helpdir).glob('*.md')
    help_pages = [x.name[:-3] for x in filelist]
    help_on = st.sidebar.selectbox(label='Help On...', options=help_pages)

    return help_on


def page_work(state_container):
    '''Main page workhorse'''

    if not 'helpSessionState' not in state_container:
        state_container.helpSessionState = HelpSessionState()

    state = state_container.helpSessionState

    url_params = st.experimental_get_query_params()
    help_on = url_params.get('help_on', [''])[0]

    if not help_on:
        st.info('Nothing to show help about')
        st.stop()

    helpdir = os.path.dirname(
        find_spec('suzieq.gui.pages').loader.path) + '/help'

    page_help_on = help_sidebar(state, helpdir)

    helptext = ''
    with open(f'{helpdir}/{help_on}.md', 'r') as f:
        helptext = f.read()

    if helptext:
        st.markdown(helptext)

    st.experimental_set_query_params(
        **{'page': '_Help_', 'help_on': help_on})
